// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#ifndef __Drucker_Prager_H
#define __Drucker_Prager_H

#include "PlasticityModel.h"
#include <dolfin/la/uBLASDenseMatrix.h>
#include <dolfin/la/uBLASVector.h>

namespace dolfin
{

  class DruckerPrager : public PlasticityModel
  {

  public:

    /// Constructor
    DruckerPrager(double E, double nu, double friction_angle, double dilatancy_angle, 
                  double cohesion, double hardening_parameter);

    /// Hardening parameter
    double hardening_parameter(double const equivalent_plastic_strain) const;

    /// Value of yield function f
    double f(const uBLASVector& stress, const double equivalent_plastic_strain);

    /// First derivative of f with respect to sigma
    void df(uBLASVector& df_dsigma, const uBLASVector& stress);

    /// First derivative of g with respect to sigma
    void dg(uBLASVector& dg_dsigma, const uBLASVector& stress);

    /// Second derivative of g with respect to sigma
    void ddg(uBLASDenseMatrix& ddg_ddsigma, const uBLASVector& stress);

  private:

    /// Computes effective stresses
    double effective_stress(const uBLASVector& stress);

    /// First derivative of g with respect to sigma, excluding (+ alpha_dilatancy/3.0) on diagonal terms
    void dg_mod(uBLASVector& dg_dsigma_mod, const uBLASVector& stress);

    /// Model parameters
    double _effective_stress, _hardening_parameter;
    double alpha_friction, k_friction, alpha_dilatancy;

    /// Auxiliary variables
    uBLASDenseMatrix A;
    uBLASVector dg_dsigma_mod;
  };
}

#endif
