// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#include "PlasticityModel.h"

using namespace dolfin;

//-----------------------------------------------------------------------------
PlasticityModel::PlasticityModel(const double E, const double nu) : _hardening_parameter(0.0)
{
  // Lame coefficients
  double lambda = nu * E / ((1 + nu) * (1 - 2 * nu));
  double mu = E / (2 * (1 + nu));

  // Create elastic tangent
  elastic_tangent.mat().resize(6, 6, false);
  elastic_tangent.mat().clear();
  elastic_tangent.mat()(0, 0) = lambda + 2 * mu;
  elastic_tangent.mat()(1, 1) = lambda + 2 * mu; 
  elastic_tangent.mat()(2, 2) = lambda + 2 * mu;
  elastic_tangent.mat()(3, 3) = mu;
  elastic_tangent.mat()(4, 4) = mu;
  elastic_tangent.mat()(5, 5) = mu;
  elastic_tangent.mat()(0, 1) = lambda;
  elastic_tangent.mat()(0, 2) = lambda;
  elastic_tangent.mat()(1, 0) = lambda;
  elastic_tangent.mat()(1, 2) = lambda;
  elastic_tangent.mat()(2, 0) = lambda;
  elastic_tangent.mat()(2, 1) = lambda;
}
//-----------------------------------------------------------------------------
PlasticityModel::~PlasticityModel()
{
  // Do nothing
}
//-----------------------------------------------------------------------------
double PlasticityModel::hardening_parameter(double const equivalent_plastic_strain) const 
{
  return _hardening_parameter;
}
//-----------------------------------------------------------------------------
double PlasticityModel::kappa(double equivalent_plastic_strain, 
                const uBLASVector& stress, const double lambda_dot) const
{
  return equivalent_plastic_strain += lambda_dot;
}
//-----------------------------------------------------------------------------
void PlasticityModel::dg(uBLASVector &dg_dsigma, const uBLASVector& stress)
{
  // Assume associative flow (dg/dsigma = df/dsigma)
  df(dg_dsigma, stress);
}
//-----------------------------------------------------------------------------
