// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#ifndef __PLASTICITY_MODEL_H
#define __PLASTICITY_MODEL_H

#include <dolfin/la/uBLASDenseMatrix.h>
#include <dolfin/la/uBLASVector.h>
#include <dolfin/la/ublas.h>

namespace dolfin
{

  class PlasticityModel
  {
  public:

    /// Constructor
    PlasticityModel(const double E, const double nu);
    
    /// Destructor
    virtual ~PlasticityModel();

    /// Hardening parameter
    virtual double hardening_parameter(double const equivalent_plastic_strain) const;

    /// Equivalent plastic strain
    virtual double kappa(double equivalent_plastic_strain, 
                       const uBLASVector& stress, const double lambda_dot) const;

    /// Value of yield function f
    virtual double f(const uBLASVector& stress, const double equivalent_plastic_strain) = 0;

    /// First derivative of f with respect to sigma
    virtual void df(uBLASVector& df_dsigma, const uBLASVector& stress) = 0;

    /// First derivative of g with respect to sigma
    virtual void dg(uBLASVector& dg_dsigma, const uBLASVector& stress);
    
    /// Second derivative of g with respect to sigma
    virtual void ddg(uBLASDenseMatrix& ddg_ddsigma, const uBLASVector& stress) = 0;
    
    friend class ReturnMapping;
    friend class PlasticityProblem;

  private:

    /// Model parameters
    double _hardening_parameter;
    uBLASDenseMatrix elastic_tangent;
  };
}

#endif
