// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#ifndef __RETURN_MAPPING_H
#define __RETURN_MAPPING_H

#include <dolfin/parameter/Parametrized.h>
#include <dolfin/la/uBLASDenseMatrix.h>
#include <dolfin/la/uBLASVector.h>
#include <dolfin/la/Vector.h>
#include <dolfin/la/GenericVector.h>
#include <dolfin/la/ublas.h>
#include "PlasticityModel.h"

namespace dolfin
{

  class ReturnMapping : public Parametrized
  {
  public:

    /// Constructor
    ReturnMapping(const uint maxit = 50);

    /// Destructor
    ~ReturnMapping();

    /// Initialise variables
    void initialise();    

    /// Closest point projection return mapping
    dolfin::uint ClosestPoint(PlasticityModel& plastic_model, uBLASDenseMatrix& consistent_tangent,
                      uBLASVector& trial_stresses, uBLASVector& plastic_strain, double& equivalent_plastic_strain);

  private:

    /// Variables for return mapping
    uBLASVector df_dsigma, dg_dsigma, sigma_current, sigma_dot, sigma_residual;
    double residual_f, delta_lambda, lambda_dot, hardening_parameter;

    /// Auxiliary variables to speed up return mapping
    uBLASDenseMatrix R;
    uBLASDenseMatrix ddg_ddsigma;
    uBLASDenseMatrix inverse_Q;
    ublas::identity_matrix<double> I;
    uBLASVector Rm, Rn, RinvQ;

    const uint maxit;
    uint plastic_flag;
  };
}

#endif
