// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#ifndef __VON_MISES_H
#define __VON_MISES_H

#include "PlasticityModel.h"
#include <dolfin/la/uBLASDenseMatrix.h>
#include <dolfin/la/uBLASVector.h>

namespace dolfin
{

  class VonMises : public PlasticityModel
  {
  public:

    /// Constructor
    VonMises(double E, double nu, double yield_stress, double hardening_parameter);

    /// Hardening parameter
    double hardening_parameter(const double equivalent_plastic_strain) const;

    /// Value of yield function f
    double f(const uBLASVector& stress, const double equivalent_plastic_strain);

    /// First derivative of f with respect to sigma
    void df(uBLASVector& df_dsigma, const uBLASVector& stress);

    /// Second derivative of g with respect to sigma
    void ddg(uBLASDenseMatrix& ddg_ddsigma, const uBLASVector& stress);

  private:

    /// Computes effective stresses
    double effective_stress(const uBLASVector& stress);

    /// Model parameters
    double _yield_stress, _effective_stress, _hardening_parameter;

    /// Auxiliary variables
    uBLASDenseMatrix A;
    uBLASVector dg_dsigma;
  };
}
#endif
