#include "Plas3D.h"
/// Constructor
UFC_Plas3DBilinearForm_finite_element_0_0::UFC_Plas3DBilinearForm_finite_element_0_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_0_0::~UFC_Plas3DBilinearForm_finite_element_0_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_0_0::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_0_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_0_0::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_0_0::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_0_0::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_0_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_0_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_0_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_0_0::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_0_0::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_0_0();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_0_1::UFC_Plas3DBilinearForm_finite_element_0_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_0_1::~UFC_Plas3DBilinearForm_finite_element_0_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_0_1::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_0_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_0_1::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_0_1::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_0_1::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_0_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_0_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_0_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_0_1::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_0_1::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_0_1();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_0_2::UFC_Plas3DBilinearForm_finite_element_0_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_0_2::~UFC_Plas3DBilinearForm_finite_element_0_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_0_2::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_0_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_0_2::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_0_2::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_0_2::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_0_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_0_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_0_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_0_2::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_0_2::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_0_2();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_0::UFC_Plas3DBilinearForm_finite_element_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_0::~UFC_Plas3DBilinearForm_finite_element_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_0::signature() const
{
    return "Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_0::space_dimension() const
{
    return 12;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_0::value_rank() const
{
    return 1;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_0::value_dimension(unsigned int i) const
{
    return 3;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[12][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[12][1] = {{1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}};
    const static double D[12][1][3] = {{{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[3];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 3; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[3] = dof_values[1];
    vertex_values[6] = dof_values[2];
    vertex_values[9] = dof_values[3];
    // Evaluate at vertices and use affine mapping
    vertex_values[1] = dof_values[4];
    vertex_values[4] = dof_values[5];
    vertex_values[7] = dof_values[6];
    vertex_values[10] = dof_values[7];
    // Evaluate at vertices and use affine mapping
    vertex_values[2] = dof_values[8];
    vertex_values[5] = dof_values[9];
    vertex_values[8] = dof_values[10];
    vertex_values[11] = dof_values[11];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_0::num_sub_elements() const
{
    return 3;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_0::create_sub_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_finite_element_0_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_finite_element_0_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_finite_element_0_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_1_0::UFC_Plas3DBilinearForm_finite_element_1_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_1_0::~UFC_Plas3DBilinearForm_finite_element_1_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_1_0::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_1_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_1_0::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_1_0::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_1_0::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_1_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_1_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_1_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_1_0::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_1_0::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_1_0();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_1_1::UFC_Plas3DBilinearForm_finite_element_1_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_1_1::~UFC_Plas3DBilinearForm_finite_element_1_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_1_1::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_1_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_1_1::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_1_1::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_1_1::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_1_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_1_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_1_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_1_1::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_1_1::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_1_1();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_1_2::UFC_Plas3DBilinearForm_finite_element_1_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_1_2::~UFC_Plas3DBilinearForm_finite_element_1_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_1_2::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_1_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_1_2::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_1_2::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_1_2::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_1_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_1_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_1_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_1_2::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_1_2::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_1_2();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_1::UFC_Plas3DBilinearForm_finite_element_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_1::~UFC_Plas3DBilinearForm_finite_element_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_1::signature() const
{
    return "Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_1::space_dimension() const
{
    return 12;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_1::value_rank() const
{
    return 1;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_1::value_dimension(unsigned int i) const
{
    return 3;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[12][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[12][1] = {{1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}};
    const static double D[12][1][3] = {{{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[3];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 3; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[3] = dof_values[1];
    vertex_values[6] = dof_values[2];
    vertex_values[9] = dof_values[3];
    // Evaluate at vertices and use affine mapping
    vertex_values[1] = dof_values[4];
    vertex_values[4] = dof_values[5];
    vertex_values[7] = dof_values[6];
    vertex_values[10] = dof_values[7];
    // Evaluate at vertices and use affine mapping
    vertex_values[2] = dof_values[8];
    vertex_values[5] = dof_values[9];
    vertex_values[8] = dof_values[10];
    vertex_values[11] = dof_values[11];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_1::num_sub_elements() const
{
    return 3;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_1::create_sub_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_finite_element_1_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_finite_element_1_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_finite_element_1_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_0::UFC_Plas3DBilinearForm_finite_element_2_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_0::~UFC_Plas3DBilinearForm_finite_element_2_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_0::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_0::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_0::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_0::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_0::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_0::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_0();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_1::UFC_Plas3DBilinearForm_finite_element_2_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_1::~UFC_Plas3DBilinearForm_finite_element_2_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_1::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_1::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_1::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_1::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_1::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_1::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_1();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_2::UFC_Plas3DBilinearForm_finite_element_2_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_2::~UFC_Plas3DBilinearForm_finite_element_2_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_2::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_2::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_2::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_2::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_2::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_2::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_2();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_3::UFC_Plas3DBilinearForm_finite_element_2_3() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_3::~UFC_Plas3DBilinearForm_finite_element_2_3()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_3::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_3::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_3::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_3::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_3::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_3::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_3::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_3::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_3::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_3::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_3::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_3::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_3::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_3::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_3();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_4::UFC_Plas3DBilinearForm_finite_element_2_4() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_4::~UFC_Plas3DBilinearForm_finite_element_2_4()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_4::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_4::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_4::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_4::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_4::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_4::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_4::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_4::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_4::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_4::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_4::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_4::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_4::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_4::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_4();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_5::UFC_Plas3DBilinearForm_finite_element_2_5() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_5::~UFC_Plas3DBilinearForm_finite_element_2_5()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_5::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_5::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_5::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_5::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_5::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_5::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_5::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_5::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_5::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_5::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_5::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_5::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_5::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_5::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_5();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_6::UFC_Plas3DBilinearForm_finite_element_2_6() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_6::~UFC_Plas3DBilinearForm_finite_element_2_6()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_6::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_6::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_6::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_6::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_6::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_6::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_6::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_6::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_6::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_6::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_6::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_6::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_6::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_6::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_6();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_7::UFC_Plas3DBilinearForm_finite_element_2_7() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_7::~UFC_Plas3DBilinearForm_finite_element_2_7()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_7::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_7::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_7::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_7::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_7::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_7::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_7::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_7::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_7::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_7::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_7::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_7::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_7::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_7::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_7();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_8::UFC_Plas3DBilinearForm_finite_element_2_8() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_8::~UFC_Plas3DBilinearForm_finite_element_2_8()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_8::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_8::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_8::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_8::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_8::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_8::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_8::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_8::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_8::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_8::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_8::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_8::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_8::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_8::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_8();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_9::UFC_Plas3DBilinearForm_finite_element_2_9() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_9::~UFC_Plas3DBilinearForm_finite_element_2_9()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_9::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_9::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_9::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_9::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_9::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_9::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_9::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_9::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_9::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_9::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_9::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_9::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_9::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_9::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_9();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_10::UFC_Plas3DBilinearForm_finite_element_2_10() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_10::~UFC_Plas3DBilinearForm_finite_element_2_10()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_10::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_10::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_10::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_10::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_10::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_10::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_10::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_10::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_10::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_10::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_10::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_10::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_10::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_10::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_10();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_11::UFC_Plas3DBilinearForm_finite_element_2_11() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_11::~UFC_Plas3DBilinearForm_finite_element_2_11()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_11::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_11::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_11::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_11::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_11::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_11::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_11::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_11::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_11::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_11::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_11::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_11::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_11::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_11::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_11();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_12::UFC_Plas3DBilinearForm_finite_element_2_12() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_12::~UFC_Plas3DBilinearForm_finite_element_2_12()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_12::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_12::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_12::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_12::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_12::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_12::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_12::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_12::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_12::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_12::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_12::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_12::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_12::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_12::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_12();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_13::UFC_Plas3DBilinearForm_finite_element_2_13() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_13::~UFC_Plas3DBilinearForm_finite_element_2_13()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_13::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_13::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_13::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_13::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_13::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_13::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_13::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_13::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_13::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_13::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_13::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_13::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_13::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_13::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_13();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_14::UFC_Plas3DBilinearForm_finite_element_2_14() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_14::~UFC_Plas3DBilinearForm_finite_element_2_14()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_14::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_14::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_14::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_14::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_14::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_14::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_14::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_14::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_14::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_14::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_14::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_14::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_14::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_14::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_14();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_15::UFC_Plas3DBilinearForm_finite_element_2_15() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_15::~UFC_Plas3DBilinearForm_finite_element_2_15()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_15::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_15::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_15::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_15::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_15::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_15::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_15::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_15::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_15::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_15::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_15::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_15::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_15::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_15::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_15();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_16::UFC_Plas3DBilinearForm_finite_element_2_16() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_16::~UFC_Plas3DBilinearForm_finite_element_2_16()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_16::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_16::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_16::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_16::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_16::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_16::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_16::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_16::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_16::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_16::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_16::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_16::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_16::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_16::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_16();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_17::UFC_Plas3DBilinearForm_finite_element_2_17() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_17::~UFC_Plas3DBilinearForm_finite_element_2_17()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_17::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_17::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_17::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_17::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_17::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_17::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_17::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_17::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_17::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_17::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_17::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_17::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_17::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_17::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_17();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_18::UFC_Plas3DBilinearForm_finite_element_2_18() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_18::~UFC_Plas3DBilinearForm_finite_element_2_18()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_18::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_18::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_18::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_18::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_18::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_18::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_18::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_18::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_18::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_18::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_18::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_18::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_18::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_18::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_18();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_19::UFC_Plas3DBilinearForm_finite_element_2_19() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_19::~UFC_Plas3DBilinearForm_finite_element_2_19()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_19::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_19::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_19::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_19::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_19::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_19::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_19::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_19::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_19::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_19::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_19::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_19::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_19::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_19::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_19();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_20::UFC_Plas3DBilinearForm_finite_element_2_20() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_20::~UFC_Plas3DBilinearForm_finite_element_2_20()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_20::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_20::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_20::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_20::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_20::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_20::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_20::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_20::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_20::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_20::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_20::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_20::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_20::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_20::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_20();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_21::UFC_Plas3DBilinearForm_finite_element_2_21() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_21::~UFC_Plas3DBilinearForm_finite_element_2_21()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_21::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_21::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_21::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_21::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_21::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_21::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_21::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_21::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_21::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_21::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_21::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_21::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_21::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_21::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_21();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_22::UFC_Plas3DBilinearForm_finite_element_2_22() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_22::~UFC_Plas3DBilinearForm_finite_element_2_22()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_22::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_22::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_22::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_22::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_22::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_22::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_22::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_22::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_22::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_22::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_22::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_22::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_22::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_22::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_22();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_23::UFC_Plas3DBilinearForm_finite_element_2_23() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_23::~UFC_Plas3DBilinearForm_finite_element_2_23()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_23::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_23::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_23::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_23::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_23::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_23::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_23::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_23::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_23::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_23::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_23::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_23::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_23::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_23::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_23();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_24::UFC_Plas3DBilinearForm_finite_element_2_24() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_24::~UFC_Plas3DBilinearForm_finite_element_2_24()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_24::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_24::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_24::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_24::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_24::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_24::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_24::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_24::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_24::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_24::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_24::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_24::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_24::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_24::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_24();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_25::UFC_Plas3DBilinearForm_finite_element_2_25() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_25::~UFC_Plas3DBilinearForm_finite_element_2_25()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_25::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_25::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_25::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_25::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_25::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_25::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_25::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_25::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_25::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_25::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_25::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_25::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_25::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_25::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_25();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_26::UFC_Plas3DBilinearForm_finite_element_2_26() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_26::~UFC_Plas3DBilinearForm_finite_element_2_26()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_26::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_26::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_26::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_26::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_26::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_26::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_26::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_26::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_26::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_26::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_26::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_26::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_26::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_26::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_26();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_27::UFC_Plas3DBilinearForm_finite_element_2_27() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_27::~UFC_Plas3DBilinearForm_finite_element_2_27()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_27::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_27::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_27::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_27::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_27::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_27::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_27::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_27::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_27::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_27::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_27::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_27::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_27::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_27::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_27();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_28::UFC_Plas3DBilinearForm_finite_element_2_28() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_28::~UFC_Plas3DBilinearForm_finite_element_2_28()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_28::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_28::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_28::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_28::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_28::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_28::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_28::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_28::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_28::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_28::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_28::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_28::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_28::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_28::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_28();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_29::UFC_Plas3DBilinearForm_finite_element_2_29() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_29::~UFC_Plas3DBilinearForm_finite_element_2_29()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_29::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_29::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_29::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_29::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_29::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_29::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_29::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_29::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_29::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_29::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_29::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_29::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_29::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_29::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_29();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_30::UFC_Plas3DBilinearForm_finite_element_2_30() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_30::~UFC_Plas3DBilinearForm_finite_element_2_30()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_30::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_30::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_30::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_30::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_30::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_30::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_30::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_30::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_30::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_30::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_30::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_30::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_30::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_30::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_30();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_31::UFC_Plas3DBilinearForm_finite_element_2_31() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_31::~UFC_Plas3DBilinearForm_finite_element_2_31()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_31::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_31::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_31::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_31::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_31::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_31::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_31::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_31::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_31::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_31::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_31::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_31::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_31::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_31::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_31();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_32::UFC_Plas3DBilinearForm_finite_element_2_32() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_32::~UFC_Plas3DBilinearForm_finite_element_2_32()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_32::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_32::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_32::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_32::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_32::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_32::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_32::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_32::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_32::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_32::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_32::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_32::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_32::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_32::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_32();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_33::UFC_Plas3DBilinearForm_finite_element_2_33() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_33::~UFC_Plas3DBilinearForm_finite_element_2_33()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_33::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_33::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_33::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_33::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_33::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_33::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_33::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_33::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_33::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_33::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_33::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_33::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_33::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_33::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_33();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_34::UFC_Plas3DBilinearForm_finite_element_2_34() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_34::~UFC_Plas3DBilinearForm_finite_element_2_34()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_34::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_34::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_34::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_34::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_34::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_34::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_34::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_34::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_34::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_34::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_34::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_34::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_34::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_34::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_34();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2_35::UFC_Plas3DBilinearForm_finite_element_2_35() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2_35::~UFC_Plas3DBilinearForm_finite_element_2_35()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2_35::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2_35::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_35::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2_35::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2_35::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_35::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_35::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_35::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2_35::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2_35::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2_35::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2_35::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2_35::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2_35::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_finite_element_2_35();
}


/// Constructor
UFC_Plas3DBilinearForm_finite_element_2::UFC_Plas3DBilinearForm_finite_element_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_finite_element_2::~UFC_Plas3DBilinearForm_finite_element_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DBilinearForm_finite_element_2::signature() const
{
    return "Mixed finite element: [Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron]";
}

/// Return the cell shape
ufc::shape UFC_Plas3DBilinearForm_finite_element_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DBilinearForm_finite_element_2::space_dimension() const
{
    return 36;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DBilinearForm_finite_element_2::value_rank() const
{
    return 1;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DBilinearForm_finite_element_2::value_dimension(unsigned int i) const
{
    return 36;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DBilinearForm_finite_element_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DBilinearForm_finite_element_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[36][1][3] = {{{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}};
    const static double W[36][1] = {{1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}};
    const static double D[36][1][36] = {{{1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0}}, {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[36];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 36; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DBilinearForm_finite_element_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DBilinearForm_finite_element_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_finite_element_2::num_sub_elements() const
{
    return 36;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DBilinearForm_finite_element_2::create_sub_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_finite_element_2_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_finite_element_2_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_finite_element_2_2();
      break;
    case 3:
      return new UFC_Plas3DBilinearForm_finite_element_2_3();
      break;
    case 4:
      return new UFC_Plas3DBilinearForm_finite_element_2_4();
      break;
    case 5:
      return new UFC_Plas3DBilinearForm_finite_element_2_5();
      break;
    case 6:
      return new UFC_Plas3DBilinearForm_finite_element_2_6();
      break;
    case 7:
      return new UFC_Plas3DBilinearForm_finite_element_2_7();
      break;
    case 8:
      return new UFC_Plas3DBilinearForm_finite_element_2_8();
      break;
    case 9:
      return new UFC_Plas3DBilinearForm_finite_element_2_9();
      break;
    case 10:
      return new UFC_Plas3DBilinearForm_finite_element_2_10();
      break;
    case 11:
      return new UFC_Plas3DBilinearForm_finite_element_2_11();
      break;
    case 12:
      return new UFC_Plas3DBilinearForm_finite_element_2_12();
      break;
    case 13:
      return new UFC_Plas3DBilinearForm_finite_element_2_13();
      break;
    case 14:
      return new UFC_Plas3DBilinearForm_finite_element_2_14();
      break;
    case 15:
      return new UFC_Plas3DBilinearForm_finite_element_2_15();
      break;
    case 16:
      return new UFC_Plas3DBilinearForm_finite_element_2_16();
      break;
    case 17:
      return new UFC_Plas3DBilinearForm_finite_element_2_17();
      break;
    case 18:
      return new UFC_Plas3DBilinearForm_finite_element_2_18();
      break;
    case 19:
      return new UFC_Plas3DBilinearForm_finite_element_2_19();
      break;
    case 20:
      return new UFC_Plas3DBilinearForm_finite_element_2_20();
      break;
    case 21:
      return new UFC_Plas3DBilinearForm_finite_element_2_21();
      break;
    case 22:
      return new UFC_Plas3DBilinearForm_finite_element_2_22();
      break;
    case 23:
      return new UFC_Plas3DBilinearForm_finite_element_2_23();
      break;
    case 24:
      return new UFC_Plas3DBilinearForm_finite_element_2_24();
      break;
    case 25:
      return new UFC_Plas3DBilinearForm_finite_element_2_25();
      break;
    case 26:
      return new UFC_Plas3DBilinearForm_finite_element_2_26();
      break;
    case 27:
      return new UFC_Plas3DBilinearForm_finite_element_2_27();
      break;
    case 28:
      return new UFC_Plas3DBilinearForm_finite_element_2_28();
      break;
    case 29:
      return new UFC_Plas3DBilinearForm_finite_element_2_29();
      break;
    case 30:
      return new UFC_Plas3DBilinearForm_finite_element_2_30();
      break;
    case 31:
      return new UFC_Plas3DBilinearForm_finite_element_2_31();
      break;
    case 32:
      return new UFC_Plas3DBilinearForm_finite_element_2_32();
      break;
    case 33:
      return new UFC_Plas3DBilinearForm_finite_element_2_33();
      break;
    case 34:
      return new UFC_Plas3DBilinearForm_finite_element_2_34();
      break;
    case 35:
      return new UFC_Plas3DBilinearForm_finite_element_2_35();
      break;
    }
    return 0;
}

/// Constructor
UFC_Plas3DBilinearForm_dof_map_0_0::UFC_Plas3DBilinearForm_dof_map_0_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_0_0::~UFC_Plas3DBilinearForm_dof_map_0_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_0_0::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_0_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_0_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_0_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_0_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0_0::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_0_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_0_0::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_0_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_0_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_0_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_0_0::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_0_0::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_0_0();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_0_1::UFC_Plas3DBilinearForm_dof_map_0_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_0_1::~UFC_Plas3DBilinearForm_dof_map_0_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_0_1::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_0_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_0_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_0_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_0_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0_1::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_0_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_0_1::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_0_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_0_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_0_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_0_1::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_0_1::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_0_1();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_0_2::UFC_Plas3DBilinearForm_dof_map_0_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_0_2::~UFC_Plas3DBilinearForm_dof_map_0_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_0_2::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_0_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_0_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_0_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_0_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0_2::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_0_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_0_2::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_0_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_0_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_0_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_0_2::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_0_2::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_0_2();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_0::UFC_Plas3DBilinearForm_dof_map_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_0::~UFC_Plas3DBilinearForm_dof_map_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_0::signature() const
{
    return "FFC dof map for Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = 3*m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_0::local_dimension() const
{
    return 12;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_0::num_facet_dofs() const
{
    return 9;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
    unsigned int offset = m.num_entities[0];
    dofs[4] = offset + c.entity_indices[0][0];
    dofs[5] = offset + c.entity_indices[0][1];
    dofs[6] = offset + c.entity_indices[0][2];
    dofs[7] = offset + c.entity_indices[0][3];
    offset = offset + m.num_entities[0];
    dofs[8] = offset + c.entity_indices[0][0];
    dofs[9] = offset + c.entity_indices[0][1];
    dofs[10] = offset + c.entity_indices[0][2];
    dofs[11] = offset + c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 5;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 9;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 11;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 6;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 10;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
    coordinates[4][0] = x[0][0];
    coordinates[4][1] = x[0][1];
    coordinates[4][2] = x[0][2];
    coordinates[5][0] = x[1][0];
    coordinates[5][1] = x[1][1];
    coordinates[5][2] = x[1][2];
    coordinates[6][0] = x[2][0];
    coordinates[6][1] = x[2][1];
    coordinates[6][2] = x[2][2];
    coordinates[7][0] = x[3][0];
    coordinates[7][1] = x[3][1];
    coordinates[7][2] = x[3][2];
    coordinates[8][0] = x[0][0];
    coordinates[8][1] = x[0][1];
    coordinates[8][2] = x[0][2];
    coordinates[9][0] = x[1][0];
    coordinates[9][1] = x[1][1];
    coordinates[9][2] = x[1][2];
    coordinates[10][0] = x[2][0];
    coordinates[10][1] = x[2][1];
    coordinates[10][2] = x[2][2];
    coordinates[11][0] = x[3][0];
    coordinates[11][1] = x[3][1];
    coordinates[11][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_0::num_sub_dof_maps() const
{
    return 3;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_0::create_sub_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_dof_map_0_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_dof_map_0_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_dof_map_0_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_1_0::UFC_Plas3DBilinearForm_dof_map_1_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_1_0::~UFC_Plas3DBilinearForm_dof_map_1_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_1_0::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_1_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_1_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_1_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_1_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1_0::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_1_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_1_0::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_1_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_1_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_1_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_1_0::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_1_0::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_1_0();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_1_1::UFC_Plas3DBilinearForm_dof_map_1_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_1_1::~UFC_Plas3DBilinearForm_dof_map_1_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_1_1::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_1_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_1_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_1_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_1_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1_1::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_1_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_1_1::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_1_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_1_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_1_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_1_1::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_1_1::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_1_1();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_1_2::UFC_Plas3DBilinearForm_dof_map_1_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_1_2::~UFC_Plas3DBilinearForm_dof_map_1_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_1_2::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_1_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_1_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_1_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_1_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1_2::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_1_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_1_2::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_1_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_1_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_1_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_1_2::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_1_2::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_1_2();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_1::UFC_Plas3DBilinearForm_dof_map_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_1::~UFC_Plas3DBilinearForm_dof_map_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_1::signature() const
{
    return "FFC dof map for Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = 3*m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_1::local_dimension() const
{
    return 12;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_1::num_facet_dofs() const
{
    return 9;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
    unsigned int offset = m.num_entities[0];
    dofs[4] = offset + c.entity_indices[0][0];
    dofs[5] = offset + c.entity_indices[0][1];
    dofs[6] = offset + c.entity_indices[0][2];
    dofs[7] = offset + c.entity_indices[0][3];
    offset = offset + m.num_entities[0];
    dofs[8] = offset + c.entity_indices[0][0];
    dofs[9] = offset + c.entity_indices[0][1];
    dofs[10] = offset + c.entity_indices[0][2];
    dofs[11] = offset + c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 5;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 9;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 11;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 6;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 10;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
    coordinates[4][0] = x[0][0];
    coordinates[4][1] = x[0][1];
    coordinates[4][2] = x[0][2];
    coordinates[5][0] = x[1][0];
    coordinates[5][1] = x[1][1];
    coordinates[5][2] = x[1][2];
    coordinates[6][0] = x[2][0];
    coordinates[6][1] = x[2][1];
    coordinates[6][2] = x[2][2];
    coordinates[7][0] = x[3][0];
    coordinates[7][1] = x[3][1];
    coordinates[7][2] = x[3][2];
    coordinates[8][0] = x[0][0];
    coordinates[8][1] = x[0][1];
    coordinates[8][2] = x[0][2];
    coordinates[9][0] = x[1][0];
    coordinates[9][1] = x[1][1];
    coordinates[9][2] = x[1][2];
    coordinates[10][0] = x[2][0];
    coordinates[10][1] = x[2][1];
    coordinates[10][2] = x[2][2];
    coordinates[11][0] = x[3][0];
    coordinates[11][1] = x[3][1];
    coordinates[11][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_1::num_sub_dof_maps() const
{
    return 3;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_1::create_sub_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_dof_map_1_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_dof_map_1_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_dof_map_1_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_0::UFC_Plas3DBilinearForm_dof_map_2_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_0::~UFC_Plas3DBilinearForm_dof_map_2_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_0::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_0::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_0::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_0::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_0::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_0();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_1::UFC_Plas3DBilinearForm_dof_map_2_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_1::~UFC_Plas3DBilinearForm_dof_map_2_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_1::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_1::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_1::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_1::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_1::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_1();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_2::UFC_Plas3DBilinearForm_dof_map_2_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_2::~UFC_Plas3DBilinearForm_dof_map_2_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_2::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_2::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_2::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_2::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_2::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_2();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_3::UFC_Plas3DBilinearForm_dof_map_2_3() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_3::~UFC_Plas3DBilinearForm_dof_map_2_3()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_3::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_3::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_3::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_3::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_3::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_3::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_3::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_3::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_3::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_3::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_3::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_3::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_3::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_3::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_3::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_3::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_3();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_4::UFC_Plas3DBilinearForm_dof_map_2_4() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_4::~UFC_Plas3DBilinearForm_dof_map_2_4()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_4::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_4::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_4::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_4::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_4::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_4::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_4::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_4::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_4::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_4::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_4::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_4::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_4::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_4::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_4::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_4::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_4();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_5::UFC_Plas3DBilinearForm_dof_map_2_5() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_5::~UFC_Plas3DBilinearForm_dof_map_2_5()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_5::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_5::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_5::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_5::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_5::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_5::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_5::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_5::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_5::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_5::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_5::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_5::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_5::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_5::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_5::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_5::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_5();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_6::UFC_Plas3DBilinearForm_dof_map_2_6() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_6::~UFC_Plas3DBilinearForm_dof_map_2_6()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_6::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_6::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_6::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_6::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_6::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_6::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_6::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_6::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_6::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_6::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_6::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_6::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_6::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_6::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_6::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_6::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_6();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_7::UFC_Plas3DBilinearForm_dof_map_2_7() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_7::~UFC_Plas3DBilinearForm_dof_map_2_7()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_7::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_7::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_7::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_7::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_7::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_7::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_7::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_7::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_7::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_7::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_7::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_7::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_7::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_7::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_7::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_7::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_7();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_8::UFC_Plas3DBilinearForm_dof_map_2_8() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_8::~UFC_Plas3DBilinearForm_dof_map_2_8()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_8::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_8::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_8::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_8::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_8::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_8::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_8::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_8::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_8::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_8::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_8::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_8::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_8::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_8::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_8::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_8::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_8();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_9::UFC_Plas3DBilinearForm_dof_map_2_9() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_9::~UFC_Plas3DBilinearForm_dof_map_2_9()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_9::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_9::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_9::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_9::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_9::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_9::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_9::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_9::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_9::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_9::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_9::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_9::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_9::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_9::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_9::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_9::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_9();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_10::UFC_Plas3DBilinearForm_dof_map_2_10() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_10::~UFC_Plas3DBilinearForm_dof_map_2_10()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_10::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_10::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_10::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_10::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_10::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_10::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_10::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_10::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_10::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_10::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_10::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_10::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_10::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_10::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_10::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_10::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_10();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_11::UFC_Plas3DBilinearForm_dof_map_2_11() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_11::~UFC_Plas3DBilinearForm_dof_map_2_11()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_11::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_11::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_11::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_11::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_11::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_11::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_11::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_11::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_11::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_11::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_11::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_11::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_11::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_11::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_11::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_11::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_11();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_12::UFC_Plas3DBilinearForm_dof_map_2_12() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_12::~UFC_Plas3DBilinearForm_dof_map_2_12()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_12::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_12::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_12::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_12::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_12::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_12::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_12::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_12::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_12::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_12::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_12::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_12::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_12::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_12::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_12::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_12::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_12();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_13::UFC_Plas3DBilinearForm_dof_map_2_13() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_13::~UFC_Plas3DBilinearForm_dof_map_2_13()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_13::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_13::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_13::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_13::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_13::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_13::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_13::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_13::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_13::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_13::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_13::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_13::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_13::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_13::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_13::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_13::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_13();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_14::UFC_Plas3DBilinearForm_dof_map_2_14() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_14::~UFC_Plas3DBilinearForm_dof_map_2_14()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_14::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_14::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_14::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_14::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_14::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_14::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_14::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_14::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_14::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_14::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_14::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_14::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_14::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_14::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_14::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_14::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_14();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_15::UFC_Plas3DBilinearForm_dof_map_2_15() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_15::~UFC_Plas3DBilinearForm_dof_map_2_15()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_15::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_15::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_15::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_15::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_15::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_15::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_15::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_15::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_15::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_15::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_15::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_15::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_15::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_15::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_15::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_15::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_15();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_16::UFC_Plas3DBilinearForm_dof_map_2_16() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_16::~UFC_Plas3DBilinearForm_dof_map_2_16()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_16::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_16::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_16::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_16::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_16::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_16::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_16::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_16::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_16::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_16::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_16::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_16::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_16::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_16::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_16::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_16::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_16();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_17::UFC_Plas3DBilinearForm_dof_map_2_17() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_17::~UFC_Plas3DBilinearForm_dof_map_2_17()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_17::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_17::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_17::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_17::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_17::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_17::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_17::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_17::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_17::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_17::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_17::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_17::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_17::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_17::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_17::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_17::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_17();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_18::UFC_Plas3DBilinearForm_dof_map_2_18() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_18::~UFC_Plas3DBilinearForm_dof_map_2_18()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_18::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_18::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_18::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_18::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_18::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_18::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_18::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_18::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_18::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_18::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_18::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_18::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_18::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_18::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_18::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_18::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_18();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_19::UFC_Plas3DBilinearForm_dof_map_2_19() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_19::~UFC_Plas3DBilinearForm_dof_map_2_19()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_19::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_19::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_19::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_19::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_19::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_19::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_19::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_19::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_19::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_19::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_19::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_19::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_19::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_19::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_19::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_19::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_19();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_20::UFC_Plas3DBilinearForm_dof_map_2_20() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_20::~UFC_Plas3DBilinearForm_dof_map_2_20()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_20::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_20::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_20::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_20::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_20::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_20::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_20::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_20::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_20::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_20::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_20::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_20::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_20::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_20::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_20::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_20::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_20();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_21::UFC_Plas3DBilinearForm_dof_map_2_21() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_21::~UFC_Plas3DBilinearForm_dof_map_2_21()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_21::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_21::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_21::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_21::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_21::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_21::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_21::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_21::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_21::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_21::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_21::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_21::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_21::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_21::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_21::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_21::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_21();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_22::UFC_Plas3DBilinearForm_dof_map_2_22() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_22::~UFC_Plas3DBilinearForm_dof_map_2_22()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_22::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_22::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_22::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_22::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_22::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_22::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_22::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_22::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_22::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_22::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_22::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_22::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_22::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_22::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_22::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_22::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_22();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_23::UFC_Plas3DBilinearForm_dof_map_2_23() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_23::~UFC_Plas3DBilinearForm_dof_map_2_23()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_23::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_23::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_23::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_23::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_23::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_23::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_23::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_23::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_23::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_23::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_23::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_23::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_23::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_23::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_23::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_23::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_23();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_24::UFC_Plas3DBilinearForm_dof_map_2_24() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_24::~UFC_Plas3DBilinearForm_dof_map_2_24()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_24::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_24::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_24::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_24::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_24::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_24::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_24::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_24::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_24::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_24::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_24::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_24::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_24::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_24::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_24::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_24::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_24();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_25::UFC_Plas3DBilinearForm_dof_map_2_25() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_25::~UFC_Plas3DBilinearForm_dof_map_2_25()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_25::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_25::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_25::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_25::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_25::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_25::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_25::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_25::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_25::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_25::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_25::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_25::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_25::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_25::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_25::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_25::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_25();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_26::UFC_Plas3DBilinearForm_dof_map_2_26() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_26::~UFC_Plas3DBilinearForm_dof_map_2_26()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_26::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_26::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_26::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_26::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_26::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_26::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_26::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_26::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_26::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_26::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_26::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_26::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_26::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_26::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_26::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_26::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_26();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_27::UFC_Plas3DBilinearForm_dof_map_2_27() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_27::~UFC_Plas3DBilinearForm_dof_map_2_27()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_27::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_27::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_27::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_27::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_27::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_27::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_27::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_27::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_27::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_27::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_27::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_27::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_27::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_27::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_27::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_27::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_27();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_28::UFC_Plas3DBilinearForm_dof_map_2_28() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_28::~UFC_Plas3DBilinearForm_dof_map_2_28()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_28::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_28::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_28::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_28::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_28::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_28::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_28::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_28::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_28::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_28::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_28::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_28::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_28::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_28::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_28::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_28::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_28();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_29::UFC_Plas3DBilinearForm_dof_map_2_29() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_29::~UFC_Plas3DBilinearForm_dof_map_2_29()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_29::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_29::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_29::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_29::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_29::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_29::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_29::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_29::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_29::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_29::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_29::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_29::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_29::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_29::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_29::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_29::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_29();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_30::UFC_Plas3DBilinearForm_dof_map_2_30() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_30::~UFC_Plas3DBilinearForm_dof_map_2_30()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_30::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_30::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_30::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_30::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_30::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_30::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_30::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_30::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_30::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_30::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_30::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_30::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_30::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_30::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_30::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_30::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_30();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_31::UFC_Plas3DBilinearForm_dof_map_2_31() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_31::~UFC_Plas3DBilinearForm_dof_map_2_31()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_31::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_31::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_31::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_31::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_31::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_31::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_31::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_31::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_31::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_31::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_31::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_31::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_31::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_31::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_31::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_31::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_31();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_32::UFC_Plas3DBilinearForm_dof_map_2_32() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_32::~UFC_Plas3DBilinearForm_dof_map_2_32()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_32::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_32::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_32::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_32::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_32::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_32::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_32::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_32::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_32::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_32::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_32::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_32::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_32::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_32::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_32::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_32::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_32();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_33::UFC_Plas3DBilinearForm_dof_map_2_33() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_33::~UFC_Plas3DBilinearForm_dof_map_2_33()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_33::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_33::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_33::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_33::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_33::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_33::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_33::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_33::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_33::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_33::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_33::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_33::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_33::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_33::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_33::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_33::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_33();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_34::UFC_Plas3DBilinearForm_dof_map_2_34() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_34::~UFC_Plas3DBilinearForm_dof_map_2_34()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_34::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_34::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_34::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_34::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_34::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_34::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_34::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_34::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_34::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_34::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_34::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_34::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_34::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_34::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_34::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_34::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_34();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2_35::UFC_Plas3DBilinearForm_dof_map_2_35() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2_35::~UFC_Plas3DBilinearForm_dof_map_2_35()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2_35::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2_35::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2_35::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2_35::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2_35::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_35::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2_35::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2_35::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2_35::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2_35::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_35::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2_35::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2_35::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2_35::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2_35::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2_35::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_dof_map_2_35();
}


/// Constructor
UFC_Plas3DBilinearForm_dof_map_2::UFC_Plas3DBilinearForm_dof_map_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DBilinearForm_dof_map_2::~UFC_Plas3DBilinearForm_dof_map_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DBilinearForm_dof_map_2::signature() const
{
    return "FFC dof map for Mixed finite element: [Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron]";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DBilinearForm_dof_map_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DBilinearForm_dof_map_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = 36*m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DBilinearForm_dof_map_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DBilinearForm_dof_map_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DBilinearForm_dof_map_2::local_dimension() const
{
    return 36;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DBilinearForm_dof_map_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DBilinearForm_dof_map_2::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DBilinearForm_dof_map_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
    unsigned int offset = m.num_entities[3];
    dofs[1] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[2] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[3] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[4] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[5] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[6] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[7] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[8] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[9] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[10] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[11] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[12] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[13] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[14] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[15] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[16] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[17] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[18] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[19] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[20] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[21] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[22] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[23] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[24] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[25] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[26] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[27] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[28] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[29] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[30] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[31] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[32] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[33] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[34] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[35] = offset + c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DBilinearForm_dof_map_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DBilinearForm_dof_map_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DBilinearForm_dof_map_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[1][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[1][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[1][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[2][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[2][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[2][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[3][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[3][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[3][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[4][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[4][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[4][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[5][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[5][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[5][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[6][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[6][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[6][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[7][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[7][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[7][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[8][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[8][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[8][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[9][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[9][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[9][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[10][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[10][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[10][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[11][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[11][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[11][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[12][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[12][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[12][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[13][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[13][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[13][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[14][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[14][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[14][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[15][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[15][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[15][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[16][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[16][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[16][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[17][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[17][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[17][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[18][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[18][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[18][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[19][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[19][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[19][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[20][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[20][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[20][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[21][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[21][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[21][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[22][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[22][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[22][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[23][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[23][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[23][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[24][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[24][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[24][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[25][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[25][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[25][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[26][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[26][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[26][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[27][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[27][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[27][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[28][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[28][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[28][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[29][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[29][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[29][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[30][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[30][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[30][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[31][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[31][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[31][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[32][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[32][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[32][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[33][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[33][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[33][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[34][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[34][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[34][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[35][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[35][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[35][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DBilinearForm_dof_map_2::num_sub_dof_maps() const
{
    return 36;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DBilinearForm_dof_map_2::create_sub_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_dof_map_2_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_dof_map_2_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_dof_map_2_2();
      break;
    case 3:
      return new UFC_Plas3DBilinearForm_dof_map_2_3();
      break;
    case 4:
      return new UFC_Plas3DBilinearForm_dof_map_2_4();
      break;
    case 5:
      return new UFC_Plas3DBilinearForm_dof_map_2_5();
      break;
    case 6:
      return new UFC_Plas3DBilinearForm_dof_map_2_6();
      break;
    case 7:
      return new UFC_Plas3DBilinearForm_dof_map_2_7();
      break;
    case 8:
      return new UFC_Plas3DBilinearForm_dof_map_2_8();
      break;
    case 9:
      return new UFC_Plas3DBilinearForm_dof_map_2_9();
      break;
    case 10:
      return new UFC_Plas3DBilinearForm_dof_map_2_10();
      break;
    case 11:
      return new UFC_Plas3DBilinearForm_dof_map_2_11();
      break;
    case 12:
      return new UFC_Plas3DBilinearForm_dof_map_2_12();
      break;
    case 13:
      return new UFC_Plas3DBilinearForm_dof_map_2_13();
      break;
    case 14:
      return new UFC_Plas3DBilinearForm_dof_map_2_14();
      break;
    case 15:
      return new UFC_Plas3DBilinearForm_dof_map_2_15();
      break;
    case 16:
      return new UFC_Plas3DBilinearForm_dof_map_2_16();
      break;
    case 17:
      return new UFC_Plas3DBilinearForm_dof_map_2_17();
      break;
    case 18:
      return new UFC_Plas3DBilinearForm_dof_map_2_18();
      break;
    case 19:
      return new UFC_Plas3DBilinearForm_dof_map_2_19();
      break;
    case 20:
      return new UFC_Plas3DBilinearForm_dof_map_2_20();
      break;
    case 21:
      return new UFC_Plas3DBilinearForm_dof_map_2_21();
      break;
    case 22:
      return new UFC_Plas3DBilinearForm_dof_map_2_22();
      break;
    case 23:
      return new UFC_Plas3DBilinearForm_dof_map_2_23();
      break;
    case 24:
      return new UFC_Plas3DBilinearForm_dof_map_2_24();
      break;
    case 25:
      return new UFC_Plas3DBilinearForm_dof_map_2_25();
      break;
    case 26:
      return new UFC_Plas3DBilinearForm_dof_map_2_26();
      break;
    case 27:
      return new UFC_Plas3DBilinearForm_dof_map_2_27();
      break;
    case 28:
      return new UFC_Plas3DBilinearForm_dof_map_2_28();
      break;
    case 29:
      return new UFC_Plas3DBilinearForm_dof_map_2_29();
      break;
    case 30:
      return new UFC_Plas3DBilinearForm_dof_map_2_30();
      break;
    case 31:
      return new UFC_Plas3DBilinearForm_dof_map_2_31();
      break;
    case 32:
      return new UFC_Plas3DBilinearForm_dof_map_2_32();
      break;
    case 33:
      return new UFC_Plas3DBilinearForm_dof_map_2_33();
      break;
    case 34:
      return new UFC_Plas3DBilinearForm_dof_map_2_34();
      break;
    case 35:
      return new UFC_Plas3DBilinearForm_dof_map_2_35();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DBilinearForm_cell_integral_0::UFC_Plas3DBilinearForm_cell_integral_0() : ufc::cell_integral()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm_cell_integral_0::~UFC_Plas3DBilinearForm_cell_integral_0()
{
    // Do nothing
}

/// Tabulate the tensor for the contribution from a local cell
void UFC_Plas3DBilinearForm_cell_integral_0::tabulate_tensor(double* A,
                                    const double * const * w,
                                    const ufc::cell& c) const
{
    // Extract vertex coordinates
    const double * const * x = c.coordinates;
    
    // Compute Jacobian of affine map from reference cell
    const double J_00 = x[1][0] - x[0][0];
    const double J_01 = x[2][0] - x[0][0];
    const double J_02 = x[3][0] - x[0][0];
    const double J_10 = x[1][1] - x[0][1];
    const double J_11 = x[2][1] - x[0][1];
    const double J_12 = x[3][1] - x[0][1];
    const double J_20 = x[1][2] - x[0][2];
    const double J_21 = x[2][2] - x[0][2];
    const double J_22 = x[3][2] - x[0][2];
      
    // Compute sub determinants
    const double d_00 = J_11*J_22 - J_12*J_21;
    const double d_01 = J_12*J_20 - J_10*J_22;
    const double d_02 = J_10*J_21 - J_11*J_20;
    
    const double d_10 = J_02*J_21 - J_01*J_22;
    const double d_11 = J_00*J_22 - J_02*J_20;
    const double d_12 = J_01*J_20 - J_00*J_21;
    
    const double d_20 = J_01*J_12 - J_02*J_11;
    const double d_21 = J_02*J_10 - J_00*J_12;
    const double d_22 = J_00*J_11 - J_01*J_10;
      
    // Compute determinant of Jacobian
    double detJ = J_00*d_00 + J_10*d_10 + J_20*d_20;
      
    // Compute inverse of Jacobian
    const double Jinv_00 = d_00 / detJ;
    const double Jinv_01 = d_10 / detJ;
    const double Jinv_02 = d_20 / detJ;
    const double Jinv_10 = d_01 / detJ;
    const double Jinv_11 = d_11 / detJ;
    const double Jinv_12 = d_21 / detJ;
    const double Jinv_20 = d_02 / detJ;
    const double Jinv_21 = d_12 / detJ;
    const double Jinv_22 = d_22 / detJ;
    
    // Set scale factor
    const double det = std::abs(detJ);
    
    
    // Compute element tensor (using quadrature representation, optimisation level 2)
    // Total number of operations to compute element tensor (from this point): 4104
    
    // Reset values of the element tensor block
    for (unsigned int j = 0; j < 12; j++)
    {
      for (unsigned int k = 0; k < 12; k++)
      {
        A[j*12 + k] = 0;
      }// end loop over 'k'
    }// end loop over 'j'
    
    // Array of quadrature weights (tensor/monomial terms (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80))
    const static double W0 = 0.166666666666666;
    
    const static double P_t2_p1_s2[1][2] = \
    {{-1, 1}};
    // Array of non-zero columns
    static const unsigned int nzc3[2] = {0, 3};
    // Array of non-zero columns
    static const unsigned int nzc4[2] = {4, 7};
    // Array of non-zero columns
    static const unsigned int nzc5[2] = {0, 1};
    // Array of non-zero columns
    static const unsigned int nzc6[2] = {0, 2};
    // Array of non-zero columns
    static const unsigned int nzc0[2] = {8, 11};
    // Array of non-zero columns
    static const unsigned int nzc2[2] = {8, 9};
    // Array of non-zero columns
    static const unsigned int nzc1[2] = {8, 10};
    // Array of non-zero columns
    static const unsigned int nzc7[2] = {4, 5};
    // Array of non-zero columns
    static const unsigned int nzc8[2] = {4, 6};
    
    // Number of operations to compute geometry constants = 2484
    const double G0 = Jinv_10*Jinv_20*W0*det*w[0][28];
    const double G1 = Jinv_10*Jinv_21*W0*det*w[0][29];
    const double G2 = Jinv_10*Jinv_22*W0*det*w[0][26];
    const double G3 = Jinv_11*Jinv_20*W0*det*w[0][34];
    const double G4 = Jinv_11*Jinv_21*W0*det*w[0][35];
    const double G5 = Jinv_11*Jinv_22*W0*det*w[0][32];
    const double G6 = Jinv_12*Jinv_20*W0*det*w[0][16];
    const double G7 = Jinv_12*Jinv_21*W0*det*w[0][17];
    const double G8 = Jinv_12*Jinv_22*W0*det*w[0][14];
    const double G9 = Jinv_10*Jinv_20*W0*det*w[0][24];
    const double G10 = Jinv_10*Jinv_21*W0*det*w[0][30];
    const double G11 = Jinv_10*Jinv_22*W0*det*w[0][12];
    const double G12 = Jinv_11*Jinv_20*W0*det*w[0][27];
    const double G13 = Jinv_11*Jinv_21*W0*det*w[0][33];
    const double G14 = Jinv_11*Jinv_22*W0*det*w[0][15];
    const double G15 = Jinv_12*Jinv_20*W0*det*w[0][28];
    const double G16 = Jinv_12*Jinv_21*W0*det*w[0][34];
    const double G17 = Jinv_12*Jinv_22*W0*det*w[0][16];
    const double G18 = Jinv_00*Jinv_00*W0*det*w[0][28];
    const double G19 = Jinv_00*Jinv_01*W0*det*w[0][29];
    const double G20 = Jinv_00*Jinv_01*W0*det*w[0][34];
    const double G21 = Jinv_00*Jinv_02*W0*det*w[0][16];
    const double G22 = Jinv_00*Jinv_02*W0*det*w[0][26];
    const double G23 = Jinv_01*Jinv_01*W0*det*w[0][35];
    const double G24 = Jinv_01*Jinv_02*W0*det*w[0][17];
    const double G25 = Jinv_01*Jinv_02*W0*det*w[0][32];
    const double G26 = Jinv_02*Jinv_02*W0*det*w[0][14];
    const double G27 = Jinv_00*Jinv_10*W0*det*w[0][27];
    const double G28 = Jinv_00*Jinv_11*W0*det*w[0][25];
    const double G29 = Jinv_00*Jinv_12*W0*det*w[0][29];
    const double G30 = Jinv_01*Jinv_10*W0*det*w[0][33];
    const double G31 = Jinv_01*Jinv_11*W0*det*w[0][31];
    const double G32 = Jinv_01*Jinv_12*W0*det*w[0][35];
    const double G33 = Jinv_02*Jinv_10*W0*det*w[0][15];
    const double G34 = Jinv_02*Jinv_11*W0*det*w[0][13];
    const double G35 = Jinv_02*Jinv_12*W0*det*w[0][17];
    const double G36 = Jinv_20*Jinv_20*W0*det*w[0][28];
    const double G37 = Jinv_20*Jinv_21*W0*det*w[0][29];
    const double G38 = Jinv_20*Jinv_21*W0*det*w[0][34];
    const double G39 = Jinv_20*Jinv_22*W0*det*w[0][16];
    const double G40 = Jinv_20*Jinv_22*W0*det*w[0][26];
    const double G41 = Jinv_21*Jinv_21*W0*det*w[0][35];
    const double G42 = Jinv_21*Jinv_22*W0*det*w[0][17];
    const double G43 = Jinv_21*Jinv_22*W0*det*w[0][32];
    const double G44 = Jinv_22*Jinv_22*W0*det*w[0][14];
    const double G45 = Jinv_10*Jinv_20*W0*det*w[0][21];
    const double G46 = Jinv_10*Jinv_21*W0*det*w[0][19];
    const double G47 = Jinv_10*Jinv_22*W0*det*w[0][23];
    const double G48 = Jinv_11*Jinv_20*W0*det*w[0][9];
    const double G49 = Jinv_11*Jinv_21*W0*det*w[0][7];
    const double G50 = Jinv_11*Jinv_22*W0*det*w[0][11];
    const double G51 = Jinv_12*Jinv_20*W0*det*w[0][33];
    const double G52 = Jinv_12*Jinv_21*W0*det*w[0][31];
    const double G53 = Jinv_12*Jinv_22*W0*det*w[0][35];
    const double G54 = Jinv_00*Jinv_20*W0*det*w[0][18];
    const double G55 = Jinv_00*Jinv_21*W0*det*w[0][21];
    const double G56 = Jinv_00*Jinv_22*W0*det*w[0][22];
    const double G57 = Jinv_01*Jinv_20*W0*det*w[0][6];
    const double G58 = Jinv_01*Jinv_21*W0*det*w[0][9];
    const double G59 = Jinv_01*Jinv_22*W0*det*w[0][10];
    const double G60 = Jinv_02*Jinv_20*W0*det*w[0][30];
    const double G61 = Jinv_02*Jinv_21*W0*det*w[0][33];
    const double G62 = Jinv_02*Jinv_22*W0*det*w[0][34];
    const double G63 = Jinv_00*Jinv_20*W0*det*w[0][24];
    const double G64 = Jinv_00*Jinv_21*W0*det*w[0][30];
    const double G65 = Jinv_00*Jinv_22*W0*det*w[0][12];
    const double G66 = Jinv_01*Jinv_20*W0*det*w[0][27];
    const double G67 = Jinv_01*Jinv_21*W0*det*w[0][33];
    const double G68 = Jinv_01*Jinv_22*W0*det*w[0][15];
    const double G69 = Jinv_02*Jinv_20*W0*det*w[0][28];
    const double G70 = Jinv_02*Jinv_21*W0*det*w[0][34];
    const double G71 = Jinv_02*Jinv_22*W0*det*w[0][16];
    const double G72 = Jinv_00*Jinv_10*W0*det*w[0][3];
    const double G73 = Jinv_00*Jinv_11*W0*det*w[0][21];
    const double G74 = Jinv_00*Jinv_12*W0*det*w[0][27];
    const double G75 = Jinv_01*Jinv_10*W0*det*w[0][1];
    const double G76 = Jinv_01*Jinv_11*W0*det*w[0][19];
    const double G77 = Jinv_01*Jinv_12*W0*det*w[0][25];
    const double G78 = Jinv_02*Jinv_10*W0*det*w[0][5];
    const double G79 = Jinv_02*Jinv_11*W0*det*w[0][23];
    const double G80 = Jinv_02*Jinv_12*W0*det*w[0][29];
    const double G81 = Jinv_00*Jinv_10*W0*det*w[0][22];
    const double G82 = Jinv_00*Jinv_11*W0*det*w[0][23];
    const double G83 = Jinv_00*Jinv_12*W0*det*w[0][20];
    const double G84 = Jinv_01*Jinv_10*W0*det*w[0][10];
    const double G85 = Jinv_01*Jinv_11*W0*det*w[0][11];
    const double G86 = Jinv_01*Jinv_12*W0*det*w[0][8];
    const double G87 = Jinv_02*Jinv_10*W0*det*w[0][34];
    const double G88 = Jinv_02*Jinv_11*W0*det*w[0][35];
    const double G89 = Jinv_02*Jinv_12*W0*det*w[0][32];
    const double G90 = Jinv_00*Jinv_10*W0*det*w[0][0];
    const double G91 = Jinv_00*Jinv_11*W0*det*w[0][18];
    const double G92 = Jinv_00*Jinv_12*W0*det*w[0][24];
    const double G93 = Jinv_01*Jinv_10*W0*det*w[0][3];
    const double G94 = Jinv_01*Jinv_11*W0*det*w[0][21];
    const double G95 = Jinv_01*Jinv_12*W0*det*w[0][27];
    const double G96 = Jinv_02*Jinv_10*W0*det*w[0][4];
    const double G97 = Jinv_02*Jinv_11*W0*det*w[0][22];
    const double G98 = Jinv_02*Jinv_12*W0*det*w[0][28];
    const double G99 = Jinv_00*Jinv_10*W0*det*w[0][28];
    const double G100 = Jinv_00*Jinv_11*W0*det*w[0][29];
    const double G101 = Jinv_00*Jinv_12*W0*det*w[0][26];
    const double G102 = Jinv_01*Jinv_10*W0*det*w[0][34];
    const double G103 = Jinv_01*Jinv_11*W0*det*w[0][35];
    const double G104 = Jinv_01*Jinv_12*W0*det*w[0][32];
    const double G105 = Jinv_02*Jinv_10*W0*det*w[0][16];
    const double G106 = Jinv_02*Jinv_11*W0*det*w[0][17];
    const double G107 = Jinv_02*Jinv_12*W0*det*w[0][14];
    const double G108 = Jinv_00*Jinv_00*W0*det*w[0][3];
    const double G109 = Jinv_00*Jinv_01*W0*det*w[0][1];
    const double G110 = Jinv_00*Jinv_01*W0*det*w[0][21];
    const double G111 = Jinv_00*Jinv_02*W0*det*w[0][27];
    const double G112 = Jinv_00*Jinv_02*W0*det*w[0][5];
    const double G113 = Jinv_01*Jinv_01*W0*det*w[0][19];
    const double G114 = Jinv_01*Jinv_02*W0*det*w[0][23];
    const double G115 = Jinv_01*Jinv_02*W0*det*w[0][25];
    const double G116 = Jinv_02*Jinv_02*W0*det*w[0][29];
    const double G117 = Jinv_00*Jinv_11*W0*det*w[0][10];
    const double G118 = Jinv_00*Jinv_12*W0*det*w[0][34];
    const double G119 = Jinv_01*Jinv_10*W0*det*w[0][23];
    const double G120 = Jinv_02*Jinv_10*W0*det*w[0][20];
    const double G121 = Jinv_02*Jinv_11*W0*det*w[0][8];
    const double G122 = Jinv_00*Jinv_20*W0*det*w[0][27];
    const double G123 = Jinv_00*Jinv_21*W0*det*w[0][33];
    const double G124 = Jinv_00*Jinv_22*W0*det*w[0][15];
    const double G125 = Jinv_01*Jinv_20*W0*det*w[0][25];
    const double G126 = Jinv_01*Jinv_21*W0*det*w[0][31];
    const double G127 = Jinv_01*Jinv_22*W0*det*w[0][13];
    const double G128 = Jinv_02*Jinv_20*W0*det*w[0][29];
    const double G129 = Jinv_02*Jinv_21*W0*det*w[0][35];
    const double G130 = Jinv_02*Jinv_22*W0*det*w[0][17];
    const double G131 = Jinv_10*Jinv_20*W0*det*w[0][22];
    const double G132 = Jinv_10*Jinv_21*W0*det*w[0][10];
    const double G133 = Jinv_10*Jinv_22*W0*det*w[0][34];
    const double G134 = Jinv_11*Jinv_20*W0*det*w[0][23];
    const double G135 = Jinv_11*Jinv_21*W0*det*w[0][11];
    const double G136 = Jinv_11*Jinv_22*W0*det*w[0][35];
    const double G137 = Jinv_12*Jinv_20*W0*det*w[0][20];
    const double G138 = Jinv_12*Jinv_21*W0*det*w[0][8];
    const double G139 = Jinv_12*Jinv_22*W0*det*w[0][32];
    const double G140 = Jinv_00*Jinv_10*W0*det*w[0][21];
    const double G141 = Jinv_00*Jinv_11*W0*det*w[0][9];
    const double G142 = Jinv_00*Jinv_12*W0*det*w[0][33];
    const double G143 = Jinv_01*Jinv_10*W0*det*w[0][19];
    const double G144 = Jinv_01*Jinv_11*W0*det*w[0][7];
    const double G145 = Jinv_01*Jinv_12*W0*det*w[0][31];
    const double G146 = Jinv_02*Jinv_10*W0*det*w[0][23];
    const double G147 = Jinv_02*Jinv_11*W0*det*w[0][11];
    const double G148 = Jinv_02*Jinv_12*W0*det*w[0][35];
    const double G149 = Jinv_10*Jinv_21*W0*det*w[0][27];
    const double G150 = Jinv_10*Jinv_22*W0*det*w[0][28];
    const double G151 = Jinv_11*Jinv_20*W0*det*w[0][30];
    const double G152 = Jinv_11*Jinv_22*W0*det*w[0][34];
    const double G153 = Jinv_12*Jinv_20*W0*det*w[0][12];
    const double G154 = Jinv_12*Jinv_21*W0*det*w[0][15];
    const double G155 = Jinv_00*Jinv_00*W0*det*w[0][0];
    const double G156 = Jinv_00*Jinv_01*W0*det*w[0][18];
    const double G157 = Jinv_00*Jinv_01*W0*det*w[0][3];
    const double G158 = Jinv_00*Jinv_02*W0*det*w[0][24];
    const double G159 = Jinv_00*Jinv_02*W0*det*w[0][4];
    const double G160 = Jinv_01*Jinv_01*W0*det*w[0][21];
    const double G161 = Jinv_01*Jinv_02*W0*det*w[0][22];
    const double G162 = Jinv_01*Jinv_02*W0*det*w[0][27];
    const double G163 = Jinv_02*Jinv_02*W0*det*w[0][28];
    const double G164 = Jinv_20*Jinv_20*W0*det*w[0][0];
    const double G165 = Jinv_20*Jinv_21*W0*det*w[0][18];
    const double G166 = Jinv_20*Jinv_21*W0*det*w[0][3];
    const double G167 = Jinv_20*Jinv_22*W0*det*w[0][24];
    const double G168 = Jinv_20*Jinv_22*W0*det*w[0][4];
    const double G169 = Jinv_21*Jinv_21*W0*det*w[0][21];
    const double G170 = Jinv_21*Jinv_22*W0*det*w[0][22];
    const double G171 = Jinv_21*Jinv_22*W0*det*w[0][27];
    const double G172 = Jinv_22*Jinv_22*W0*det*w[0][28];
    const double G173 = Jinv_00*Jinv_00*W0*det*w[0][21];
    const double G174 = Jinv_00*Jinv_01*W0*det*w[0][19];
    const double G175 = Jinv_00*Jinv_01*W0*det*w[0][9];
    const double G176 = Jinv_00*Jinv_02*W0*det*w[0][23];
    const double G177 = Jinv_00*Jinv_02*W0*det*w[0][33];
    const double G178 = Jinv_01*Jinv_01*W0*det*w[0][7];
    const double G179 = Jinv_01*Jinv_02*W0*det*w[0][11];
    const double G180 = Jinv_01*Jinv_02*W0*det*w[0][31];
    const double G181 = Jinv_02*Jinv_02*W0*det*w[0][35];
    const double G182 = Jinv_20*Jinv_20*W0*det*w[0][22];
    const double G183 = Jinv_20*Jinv_21*W0*det*w[0][10];
    const double G184 = Jinv_20*Jinv_21*W0*det*w[0][23];
    const double G185 = Jinv_20*Jinv_22*W0*det*w[0][20];
    const double G186 = Jinv_20*Jinv_22*W0*det*w[0][34];
    const double G187 = Jinv_21*Jinv_21*W0*det*w[0][11];
    const double G188 = Jinv_21*Jinv_22*W0*det*w[0][35];
    const double G189 = Jinv_21*Jinv_22*W0*det*w[0][8];
    const double G190 = Jinv_22*Jinv_22*W0*det*w[0][32];
    const double G191 = Jinv_20*Jinv_20*W0*det*w[0][4];
    const double G192 = Jinv_20*Jinv_21*W0*det*w[0][22];
    const double G193 = Jinv_20*Jinv_21*W0*det*w[0][5];
    const double G194 = Jinv_20*Jinv_22*W0*det*w[0][28];
    const double G195 = Jinv_20*Jinv_22*W0*det*w[0][2];
    const double G196 = Jinv_21*Jinv_21*W0*det*w[0][23];
    const double G197 = Jinv_21*Jinv_22*W0*det*w[0][20];
    const double G198 = Jinv_21*Jinv_22*W0*det*w[0][29];
    const double G199 = Jinv_22*Jinv_22*W0*det*w[0][26];
    const double G200 = Jinv_10*Jinv_10*W0*det*w[0][28];
    const double G201 = Jinv_10*Jinv_11*W0*det*w[0][29];
    const double G202 = Jinv_10*Jinv_11*W0*det*w[0][34];
    const double G203 = Jinv_10*Jinv_12*W0*det*w[0][16];
    const double G204 = Jinv_10*Jinv_12*W0*det*w[0][26];
    const double G205 = Jinv_11*Jinv_11*W0*det*w[0][35];
    const double G206 = Jinv_11*Jinv_12*W0*det*w[0][17];
    const double G207 = Jinv_11*Jinv_12*W0*det*w[0][32];
    const double G208 = Jinv_12*Jinv_12*W0*det*w[0][14];
    const double G209 = Jinv_10*Jinv_20*W0*det*w[0][27];
    const double G210 = Jinv_10*Jinv_21*W0*det*w[0][25];
    const double G211 = Jinv_10*Jinv_22*W0*det*w[0][29];
    const double G212 = Jinv_11*Jinv_20*W0*det*w[0][33];
    const double G213 = Jinv_11*Jinv_21*W0*det*w[0][31];
    const double G214 = Jinv_12*Jinv_20*W0*det*w[0][15];
    const double G215 = Jinv_12*Jinv_21*W0*det*w[0][13];
    const double G216 = Jinv_12*Jinv_22*W0*det*w[0][17];
    const double G217 = Jinv_10*Jinv_21*W0*det*w[0][23];
    const double G218 = Jinv_10*Jinv_22*W0*det*w[0][20];
    const double G219 = Jinv_11*Jinv_20*W0*det*w[0][10];
    const double G220 = Jinv_11*Jinv_22*W0*det*w[0][8];
    const double G221 = Jinv_12*Jinv_20*W0*det*w[0][34];
    const double G222 = Jinv_12*Jinv_21*W0*det*w[0][35];
    const double G223 = Jinv_10*Jinv_21*W0*det*w[0][33];
    const double G224 = Jinv_10*Jinv_22*W0*det*w[0][15];
    const double G225 = Jinv_11*Jinv_20*W0*det*w[0][25];
    const double G226 = Jinv_11*Jinv_22*W0*det*w[0][13];
    const double G227 = Jinv_12*Jinv_20*W0*det*w[0][29];
    const double G228 = Jinv_10*Jinv_20*W0*det*w[0][3];
    const double G229 = Jinv_10*Jinv_21*W0*det*w[0][21];
    const double G230 = Jinv_10*Jinv_22*W0*det*w[0][27];
    const double G231 = Jinv_11*Jinv_20*W0*det*w[0][1];
    const double G232 = Jinv_11*Jinv_21*W0*det*w[0][19];
    const double G233 = Jinv_11*Jinv_22*W0*det*w[0][25];
    const double G234 = Jinv_12*Jinv_20*W0*det*w[0][5];
    const double G235 = Jinv_12*Jinv_21*W0*det*w[0][23];
    const double G236 = Jinv_12*Jinv_22*W0*det*w[0][29];
    const double G237 = Jinv_10*Jinv_10*W0*det*w[0][18];
    const double G238 = Jinv_10*Jinv_11*W0*det*w[0][21];
    const double G239 = Jinv_10*Jinv_11*W0*det*w[0][6];
    const double G240 = Jinv_10*Jinv_12*W0*det*w[0][22];
    const double G241 = Jinv_10*Jinv_12*W0*det*w[0][30];
    const double G242 = Jinv_11*Jinv_11*W0*det*w[0][9];
    const double G243 = Jinv_11*Jinv_12*W0*det*w[0][10];
    const double G244 = Jinv_11*Jinv_12*W0*det*w[0][33];
    const double G245 = Jinv_12*Jinv_12*W0*det*w[0][34];
    const double G246 = Jinv_00*Jinv_00*W0*det*w[0][24];
    const double G247 = Jinv_00*Jinv_01*W0*det*w[0][27];
    const double G248 = Jinv_00*Jinv_01*W0*det*w[0][30];
    const double G249 = Jinv_00*Jinv_02*W0*det*w[0][12];
    const double G250 = Jinv_00*Jinv_02*W0*det*w[0][28];
    const double G251 = Jinv_01*Jinv_01*W0*det*w[0][33];
    const double G252 = Jinv_01*Jinv_02*W0*det*w[0][15];
    const double G253 = Jinv_01*Jinv_02*W0*det*w[0][34];
    const double G254 = Jinv_02*Jinv_02*W0*det*w[0][16];
    const double G255 = Jinv_00*Jinv_11*W0*det*w[0][3];
    const double G256 = Jinv_00*Jinv_12*W0*det*w[0][4];
    const double G257 = Jinv_01*Jinv_10*W0*det*w[0][18];
    const double G258 = Jinv_01*Jinv_12*W0*det*w[0][22];
    const double G259 = Jinv_02*Jinv_10*W0*det*w[0][24];
    const double G260 = Jinv_02*Jinv_11*W0*det*w[0][27];
    const double G261 = Jinv_00*Jinv_00*W0*det*w[0][18];
    const double G262 = Jinv_00*Jinv_01*W0*det*w[0][6];
    const double G263 = Jinv_00*Jinv_02*W0*det*w[0][22];
    const double G264 = Jinv_00*Jinv_02*W0*det*w[0][30];
    const double G265 = Jinv_01*Jinv_01*W0*det*w[0][9];
    const double G266 = Jinv_01*Jinv_02*W0*det*w[0][10];
    const double G267 = Jinv_01*Jinv_02*W0*det*w[0][33];
    const double G268 = Jinv_02*Jinv_02*W0*det*w[0][34];
    const double G269 = Jinv_00*Jinv_11*W0*det*w[0][33];
    const double G270 = Jinv_00*Jinv_12*W0*det*w[0][15];
    const double G271 = Jinv_01*Jinv_10*W0*det*w[0][25];
    const double G272 = Jinv_01*Jinv_12*W0*det*w[0][13];
    const double G273 = Jinv_02*Jinv_10*W0*det*w[0][29];
    const double G274 = Jinv_10*Jinv_20*W0*det*w[0][0];
    const double G275 = Jinv_10*Jinv_21*W0*det*w[0][3];
    const double G276 = Jinv_10*Jinv_22*W0*det*w[0][4];
    const double G277 = Jinv_11*Jinv_20*W0*det*w[0][18];
    const double G278 = Jinv_11*Jinv_21*W0*det*w[0][21];
    const double G279 = Jinv_11*Jinv_22*W0*det*w[0][22];
    const double G280 = Jinv_12*Jinv_20*W0*det*w[0][24];
    const double G281 = Jinv_12*Jinv_21*W0*det*w[0][27];
    const double G282 = Jinv_12*Jinv_22*W0*det*w[0][28];
    const double G283 = Jinv_10*Jinv_10*W0*det*w[0][3];
    const double G284 = Jinv_10*Jinv_11*W0*det*w[0][1];
    const double G285 = Jinv_10*Jinv_12*W0*det*w[0][27];
    const double G286 = Jinv_10*Jinv_12*W0*det*w[0][5];
    const double G287 = Jinv_11*Jinv_11*W0*det*w[0][19];
    const double G288 = Jinv_11*Jinv_12*W0*det*w[0][23];
    const double G289 = Jinv_11*Jinv_12*W0*det*w[0][25];
    const double G290 = Jinv_12*Jinv_12*W0*det*w[0][29];
    const double G291 = Jinv_00*Jinv_11*W0*det*w[0][1];
    const double G292 = Jinv_00*Jinv_12*W0*det*w[0][5];
    const double G293 = Jinv_01*Jinv_10*W0*det*w[0][21];
    const double G294 = Jinv_01*Jinv_12*W0*det*w[0][23];
    const double G295 = Jinv_02*Jinv_10*W0*det*w[0][27];
    const double G296 = Jinv_02*Jinv_11*W0*det*w[0][25];
    const double G297 = Jinv_00*Jinv_11*W0*det*w[0][34];
    const double G298 = Jinv_00*Jinv_12*W0*det*w[0][16];
    const double G299 = Jinv_01*Jinv_10*W0*det*w[0][29];
    const double G300 = Jinv_01*Jinv_12*W0*det*w[0][17];
    const double G301 = Jinv_02*Jinv_10*W0*det*w[0][26];
    const double G302 = Jinv_02*Jinv_11*W0*det*w[0][32];
    const double G303 = Jinv_20*Jinv_20*W0*det*w[0][24];
    const double G304 = Jinv_20*Jinv_21*W0*det*w[0][27];
    const double G305 = Jinv_20*Jinv_21*W0*det*w[0][30];
    const double G306 = Jinv_20*Jinv_22*W0*det*w[0][12];
    const double G307 = Jinv_21*Jinv_21*W0*det*w[0][33];
    const double G308 = Jinv_21*Jinv_22*W0*det*w[0][15];
    const double G309 = Jinv_21*Jinv_22*W0*det*w[0][34];
    const double G310 = Jinv_22*Jinv_22*W0*det*w[0][16];
    const double G311 = Jinv_10*Jinv_21*W0*det*w[0][18];
    const double G312 = Jinv_10*Jinv_22*W0*det*w[0][24];
    const double G313 = Jinv_11*Jinv_20*W0*det*w[0][3];
    const double G314 = Jinv_11*Jinv_22*W0*det*w[0][27];
    const double G315 = Jinv_12*Jinv_20*W0*det*w[0][4];
    const double G316 = Jinv_12*Jinv_21*W0*det*w[0][22];
    const double G317 = Jinv_00*Jinv_10*W0*det*w[0][24];
    const double G318 = Jinv_00*Jinv_11*W0*det*w[0][27];
    const double G319 = Jinv_00*Jinv_12*W0*det*w[0][28];
    const double G320 = Jinv_01*Jinv_10*W0*det*w[0][30];
    const double G321 = Jinv_01*Jinv_11*W0*det*w[0][33];
    const double G322 = Jinv_01*Jinv_12*W0*det*w[0][34];
    const double G323 = Jinv_02*Jinv_10*W0*det*w[0][12];
    const double G324 = Jinv_02*Jinv_11*W0*det*w[0][15];
    const double G325 = Jinv_02*Jinv_12*W0*det*w[0][16];
    const double G326 = Jinv_00*Jinv_00*W0*det*w[0][4];
    const double G327 = Jinv_00*Jinv_01*W0*det*w[0][22];
    const double G328 = Jinv_00*Jinv_01*W0*det*w[0][5];
    const double G329 = Jinv_00*Jinv_02*W0*det*w[0][2];
    const double G330 = Jinv_01*Jinv_01*W0*det*w[0][23];
    const double G331 = Jinv_01*Jinv_02*W0*det*w[0][20];
    const double G332 = Jinv_01*Jinv_02*W0*det*w[0][29];
    const double G333 = Jinv_02*Jinv_02*W0*det*w[0][26];
    const double G334 = Jinv_00*Jinv_20*W0*det*w[0][4];
    const double G335 = Jinv_00*Jinv_21*W0*det*w[0][5];
    const double G336 = Jinv_00*Jinv_22*W0*det*w[0][2];
    const double G337 = Jinv_01*Jinv_20*W0*det*w[0][22];
    const double G338 = Jinv_01*Jinv_21*W0*det*w[0][23];
    const double G339 = Jinv_01*Jinv_22*W0*det*w[0][20];
    const double G340 = Jinv_02*Jinv_21*W0*det*w[0][29];
    const double G341 = Jinv_02*Jinv_22*W0*det*w[0][26];
    const double G342 = Jinv_10*Jinv_21*W0*det*w[0][34];
    const double G343 = Jinv_10*Jinv_22*W0*det*w[0][16];
    const double G344 = Jinv_11*Jinv_20*W0*det*w[0][29];
    const double G345 = Jinv_11*Jinv_22*W0*det*w[0][17];
    const double G346 = Jinv_12*Jinv_20*W0*det*w[0][26];
    const double G347 = Jinv_12*Jinv_21*W0*det*w[0][32];
    const double G348 = Jinv_00*Jinv_20*W0*det*w[0][3];
    const double G349 = Jinv_00*Jinv_22*W0*det*w[0][27];
    const double G350 = Jinv_01*Jinv_20*W0*det*w[0][1];
    const double G351 = Jinv_01*Jinv_21*W0*det*w[0][19];
    const double G352 = Jinv_01*Jinv_22*W0*det*w[0][25];
    const double G353 = Jinv_02*Jinv_20*W0*det*w[0][5];
    const double G354 = Jinv_02*Jinv_21*W0*det*w[0][23];
    const double G355 = Jinv_02*Jinv_22*W0*det*w[0][29];
    const double G356 = Jinv_00*Jinv_10*W0*det*w[0][4];
    const double G357 = Jinv_00*Jinv_11*W0*det*w[0][22];
    const double G358 = Jinv_01*Jinv_10*W0*det*w[0][5];
    const double G359 = Jinv_01*Jinv_11*W0*det*w[0][23];
    const double G360 = Jinv_01*Jinv_12*W0*det*w[0][29];
    const double G361 = Jinv_02*Jinv_10*W0*det*w[0][2];
    const double G362 = Jinv_02*Jinv_11*W0*det*w[0][20];
    const double G363 = Jinv_02*Jinv_12*W0*det*w[0][26];
    const double G364 = Jinv_10*Jinv_20*W0*det*w[0][18];
    const double G365 = Jinv_10*Jinv_21*W0*det*w[0][6];
    const double G366 = Jinv_10*Jinv_22*W0*det*w[0][30];
    const double G367 = Jinv_11*Jinv_20*W0*det*w[0][21];
    const double G368 = Jinv_11*Jinv_21*W0*det*w[0][9];
    const double G369 = Jinv_11*Jinv_22*W0*det*w[0][33];
    const double G370 = Jinv_12*Jinv_20*W0*det*w[0][22];
    const double G371 = Jinv_12*Jinv_21*W0*det*w[0][10];
    const double G372 = Jinv_12*Jinv_22*W0*det*w[0][34];
    const double G373 = Jinv_20*Jinv_20*W0*det*w[0][21];
    const double G374 = Jinv_20*Jinv_21*W0*det*w[0][19];
    const double G375 = Jinv_20*Jinv_21*W0*det*w[0][9];
    const double G376 = Jinv_20*Jinv_22*W0*det*w[0][23];
    const double G377 = Jinv_20*Jinv_22*W0*det*w[0][33];
    const double G378 = Jinv_21*Jinv_21*W0*det*w[0][7];
    const double G379 = Jinv_21*Jinv_22*W0*det*w[0][11];
    const double G380 = Jinv_21*Jinv_22*W0*det*w[0][31];
    const double G381 = Jinv_22*Jinv_22*W0*det*w[0][35];
    const double G382 = Jinv_10*Jinv_10*W0*det*w[0][22];
    const double G383 = Jinv_10*Jinv_11*W0*det*w[0][10];
    const double G384 = Jinv_10*Jinv_11*W0*det*w[0][23];
    const double G385 = Jinv_10*Jinv_12*W0*det*w[0][20];
    const double G386 = Jinv_10*Jinv_12*W0*det*w[0][34];
    const double G387 = Jinv_11*Jinv_11*W0*det*w[0][11];
    const double G388 = Jinv_11*Jinv_12*W0*det*w[0][35];
    const double G389 = Jinv_11*Jinv_12*W0*det*w[0][8];
    const double G390 = Jinv_12*Jinv_12*W0*det*w[0][32];
    const double G391 = Jinv_10*Jinv_22*W0*det*w[0][22];
    const double G392 = Jinv_11*Jinv_20*W0*det*w[0][6];
    const double G393 = Jinv_11*Jinv_22*W0*det*w[0][10];
    const double G394 = Jinv_12*Jinv_20*W0*det*w[0][30];
    const double G395 = Jinv_12*Jinv_21*W0*det*w[0][33];
    const double G396 = Jinv_00*Jinv_20*W0*det*w[0][0];
    const double G397 = Jinv_00*Jinv_21*W0*det*w[0][3];
    const double G398 = Jinv_00*Jinv_22*W0*det*w[0][4];
    const double G399 = Jinv_01*Jinv_20*W0*det*w[0][18];
    const double G400 = Jinv_01*Jinv_21*W0*det*w[0][21];
    const double G401 = Jinv_01*Jinv_22*W0*det*w[0][22];
    const double G402 = Jinv_02*Jinv_20*W0*det*w[0][24];
    const double G403 = Jinv_02*Jinv_21*W0*det*w[0][27];
    const double G404 = Jinv_02*Jinv_22*W0*det*w[0][28];
    const double G405 = Jinv_00*Jinv_20*W0*det*w[0][22];
    const double G406 = Jinv_00*Jinv_21*W0*det*w[0][10];
    const double G407 = Jinv_00*Jinv_22*W0*det*w[0][34];
    const double G408 = Jinv_01*Jinv_20*W0*det*w[0][23];
    const double G409 = Jinv_01*Jinv_21*W0*det*w[0][11];
    const double G410 = Jinv_01*Jinv_22*W0*det*w[0][35];
    const double G411 = Jinv_02*Jinv_20*W0*det*w[0][20];
    const double G412 = Jinv_02*Jinv_21*W0*det*w[0][8];
    const double G413 = Jinv_02*Jinv_22*W0*det*w[0][32];
    const double G414 = Jinv_00*Jinv_21*W0*det*w[0][23];
    const double G415 = Jinv_00*Jinv_22*W0*det*w[0][20];
    const double G416 = Jinv_01*Jinv_20*W0*det*w[0][10];
    const double G417 = Jinv_01*Jinv_22*W0*det*w[0][8];
    const double G418 = Jinv_02*Jinv_20*W0*det*w[0][34];
    const double G419 = Jinv_10*Jinv_21*W0*det*w[0][1];
    const double G420 = Jinv_10*Jinv_22*W0*det*w[0][5];
    const double G421 = Jinv_11*Jinv_22*W0*det*w[0][23];
    const double G422 = Jinv_12*Jinv_20*W0*det*w[0][27];
    const double G423 = Jinv_12*Jinv_21*W0*det*w[0][25];
    const double G424 = Jinv_00*Jinv_20*W0*det*w[0][28];
    const double G425 = Jinv_00*Jinv_21*W0*det*w[0][29];
    const double G426 = Jinv_00*Jinv_22*W0*det*w[0][26];
    const double G427 = Jinv_01*Jinv_20*W0*det*w[0][34];
    const double G428 = Jinv_01*Jinv_21*W0*det*w[0][35];
    const double G429 = Jinv_01*Jinv_22*W0*det*w[0][32];
    const double G430 = Jinv_02*Jinv_20*W0*det*w[0][16];
    const double G431 = Jinv_02*Jinv_21*W0*det*w[0][17];
    const double G432 = Jinv_02*Jinv_22*W0*det*w[0][14];
    const double G433 = Jinv_20*Jinv_20*W0*det*w[0][18];
    const double G434 = Jinv_20*Jinv_21*W0*det*w[0][21];
    const double G435 = Jinv_20*Jinv_21*W0*det*w[0][6];
    const double G436 = Jinv_20*Jinv_22*W0*det*w[0][22];
    const double G437 = Jinv_20*Jinv_22*W0*det*w[0][30];
    const double G438 = Jinv_21*Jinv_21*W0*det*w[0][9];
    const double G439 = Jinv_21*Jinv_22*W0*det*w[0][10];
    const double G440 = Jinv_21*Jinv_22*W0*det*w[0][33];
    const double G441 = Jinv_22*Jinv_22*W0*det*w[0][34];
    const double G442 = Jinv_00*Jinv_21*W0*det*w[0][18];
    const double G443 = Jinv_00*Jinv_22*W0*det*w[0][24];
    const double G444 = Jinv_01*Jinv_20*W0*det*w[0][3];
    const double G445 = Jinv_01*Jinv_22*W0*det*w[0][27];
    const double G446 = Jinv_02*Jinv_20*W0*det*w[0][4];
    const double G447 = Jinv_02*Jinv_21*W0*det*w[0][22];
    const double G448 = Jinv_00*Jinv_21*W0*det*w[0][34];
    const double G449 = Jinv_00*Jinv_22*W0*det*w[0][16];
    const double G450 = Jinv_01*Jinv_20*W0*det*w[0][29];
    const double G451 = Jinv_01*Jinv_22*W0*det*w[0][17];
    const double G452 = Jinv_02*Jinv_20*W0*det*w[0][26];
    const double G453 = Jinv_02*Jinv_21*W0*det*w[0][32];
    const double G454 = Jinv_00*Jinv_00*W0*det*w[0][22];
    const double G455 = Jinv_00*Jinv_01*W0*det*w[0][10];
    const double G456 = Jinv_00*Jinv_01*W0*det*w[0][23];
    const double G457 = Jinv_00*Jinv_02*W0*det*w[0][20];
    const double G458 = Jinv_00*Jinv_02*W0*det*w[0][34];
    const double G459 = Jinv_01*Jinv_01*W0*det*w[0][11];
    const double G460 = Jinv_01*Jinv_02*W0*det*w[0][35];
    const double G461 = Jinv_01*Jinv_02*W0*det*w[0][8];
    const double G462 = Jinv_02*Jinv_02*W0*det*w[0][32];
    const double G463 = Jinv_00*Jinv_11*W0*det*w[0][30];
    const double G464 = Jinv_00*Jinv_12*W0*det*w[0][12];
    const double G465 = Jinv_01*Jinv_10*W0*det*w[0][27];
    const double G466 = Jinv_01*Jinv_12*W0*det*w[0][15];
    const double G467 = Jinv_02*Jinv_10*W0*det*w[0][28];
    const double G468 = Jinv_02*Jinv_11*W0*det*w[0][34];
    const double G469 = Jinv_00*Jinv_21*W0*det*w[0][27];
    const double G470 = Jinv_00*Jinv_22*W0*det*w[0][28];
    const double G471 = Jinv_01*Jinv_20*W0*det*w[0][30];
    const double G472 = Jinv_01*Jinv_22*W0*det*w[0][34];
    const double G473 = Jinv_02*Jinv_20*W0*det*w[0][12];
    const double G474 = Jinv_02*Jinv_21*W0*det*w[0][15];
    const double G475 = Jinv_20*Jinv_20*W0*det*w[0][27];
    const double G476 = Jinv_20*Jinv_21*W0*det*w[0][25];
    const double G477 = Jinv_20*Jinv_21*W0*det*w[0][33];
    const double G478 = Jinv_20*Jinv_22*W0*det*w[0][15];
    const double G479 = Jinv_20*Jinv_22*W0*det*w[0][29];
    const double G480 = Jinv_21*Jinv_21*W0*det*w[0][31];
    const double G481 = Jinv_21*Jinv_22*W0*det*w[0][13];
    const double G482 = Jinv_22*Jinv_22*W0*det*w[0][17];
    const double G483 = Jinv_10*Jinv_10*W0*det*w[0][27];
    const double G484 = Jinv_10*Jinv_11*W0*det*w[0][25];
    const double G485 = Jinv_10*Jinv_11*W0*det*w[0][33];
    const double G486 = Jinv_10*Jinv_12*W0*det*w[0][15];
    const double G487 = Jinv_10*Jinv_12*W0*det*w[0][29];
    const double G488 = Jinv_11*Jinv_11*W0*det*w[0][31];
    const double G489 = Jinv_11*Jinv_12*W0*det*w[0][13];
    const double G490 = Jinv_12*Jinv_12*W0*det*w[0][17];
    const double G491 = Jinv_00*Jinv_10*W0*det*w[0][18];
    const double G492 = Jinv_00*Jinv_12*W0*det*w[0][22];
    const double G493 = Jinv_01*Jinv_10*W0*det*w[0][6];
    const double G494 = Jinv_01*Jinv_11*W0*det*w[0][9];
    const double G495 = Jinv_01*Jinv_12*W0*det*w[0][10];
    const double G496 = Jinv_02*Jinv_10*W0*det*w[0][30];
    const double G497 = Jinv_02*Jinv_11*W0*det*w[0][33];
    const double G498 = Jinv_02*Jinv_12*W0*det*w[0][34];
    const double G499 = Jinv_10*Jinv_10*W0*det*w[0][24];
    const double G500 = Jinv_10*Jinv_11*W0*det*w[0][27];
    const double G501 = Jinv_10*Jinv_11*W0*det*w[0][30];
    const double G502 = Jinv_10*Jinv_12*W0*det*w[0][12];
    const double G503 = Jinv_10*Jinv_12*W0*det*w[0][28];
    const double G504 = Jinv_11*Jinv_11*W0*det*w[0][33];
    const double G505 = Jinv_11*Jinv_12*W0*det*w[0][15];
    const double G506 = Jinv_11*Jinv_12*W0*det*w[0][34];
    const double G507 = Jinv_12*Jinv_12*W0*det*w[0][16];
    const double G508 = Jinv_00*Jinv_21*W0*det*w[0][25];
    const double G509 = Jinv_00*Jinv_22*W0*det*w[0][29];
    const double G510 = Jinv_01*Jinv_20*W0*det*w[0][33];
    const double G511 = Jinv_02*Jinv_20*W0*det*w[0][15];
    const double G512 = Jinv_02*Jinv_21*W0*det*w[0][13];
    const double G513 = Jinv_00*Jinv_21*W0*det*w[0][22];
    const double G514 = Jinv_01*Jinv_20*W0*det*w[0][5];
    const double G515 = Jinv_01*Jinv_22*W0*det*w[0][29];
    const double G516 = Jinv_02*Jinv_20*W0*det*w[0][2];
    const double G517 = Jinv_02*Jinv_21*W0*det*w[0][20];
    const double G518 = Jinv_10*Jinv_10*W0*det*w[0][0];
    const double G519 = Jinv_10*Jinv_11*W0*det*w[0][18];
    const double G520 = Jinv_10*Jinv_11*W0*det*w[0][3];
    const double G521 = Jinv_10*Jinv_12*W0*det*w[0][24];
    const double G522 = Jinv_10*Jinv_12*W0*det*w[0][4];
    const double G523 = Jinv_11*Jinv_11*W0*det*w[0][21];
    const double G524 = Jinv_11*Jinv_12*W0*det*w[0][22];
    const double G525 = Jinv_11*Jinv_12*W0*det*w[0][27];
    const double G526 = Jinv_12*Jinv_12*W0*det*w[0][28];
    const double G527 = Jinv_00*Jinv_21*W0*det*w[0][6];
    const double G528 = Jinv_00*Jinv_22*W0*det*w[0][30];
    const double G529 = Jinv_01*Jinv_20*W0*det*w[0][21];
    const double G530 = Jinv_01*Jinv_22*W0*det*w[0][33];
    const double G531 = Jinv_02*Jinv_20*W0*det*w[0][22];
    const double G532 = Jinv_02*Jinv_21*W0*det*w[0][10];
    const double G533 = Jinv_00*Jinv_20*W0*det*w[0][21];
    const double G534 = Jinv_00*Jinv_21*W0*det*w[0][9];
    const double G535 = Jinv_00*Jinv_22*W0*det*w[0][33];
    const double G536 = Jinv_01*Jinv_20*W0*det*w[0][19];
    const double G537 = Jinv_01*Jinv_21*W0*det*w[0][7];
    const double G538 = Jinv_01*Jinv_22*W0*det*w[0][31];
    const double G539 = Jinv_02*Jinv_20*W0*det*w[0][23];
    const double G540 = Jinv_02*Jinv_21*W0*det*w[0][11];
    const double G541 = Jinv_02*Jinv_22*W0*det*w[0][35];
    const double G542 = Jinv_10*Jinv_21*W0*det*w[0][9];
    const double G543 = Jinv_10*Jinv_22*W0*det*w[0][33];
    const double G544 = Jinv_11*Jinv_20*W0*det*w[0][19];
    const double G545 = Jinv_11*Jinv_22*W0*det*w[0][31];
    const double G546 = Jinv_12*Jinv_20*W0*det*w[0][23];
    const double G547 = Jinv_12*Jinv_21*W0*det*w[0][11];
    const double G548 = Jinv_00*Jinv_00*W0*det*w[0][27];
    const double G549 = Jinv_00*Jinv_01*W0*det*w[0][25];
    const double G550 = Jinv_00*Jinv_01*W0*det*w[0][33];
    const double G551 = Jinv_00*Jinv_02*W0*det*w[0][15];
    const double G552 = Jinv_00*Jinv_02*W0*det*w[0][29];
    const double G553 = Jinv_01*Jinv_01*W0*det*w[0][31];
    const double G554 = Jinv_01*Jinv_02*W0*det*w[0][13];
    const double G555 = Jinv_02*Jinv_02*W0*det*w[0][17];
    const double G556 = Jinv_10*Jinv_20*W0*det*w[0][4];
    const double G557 = Jinv_10*Jinv_21*W0*det*w[0][5];
    const double G558 = Jinv_10*Jinv_22*W0*det*w[0][2];
    const double G559 = Jinv_11*Jinv_20*W0*det*w[0][22];
    const double G560 = Jinv_11*Jinv_21*W0*det*w[0][23];
    const double G561 = Jinv_11*Jinv_22*W0*det*w[0][20];
    const double G562 = Jinv_12*Jinv_21*W0*det*w[0][29];
    const double G563 = Jinv_12*Jinv_22*W0*det*w[0][26];
    const double G564 = Jinv_00*Jinv_11*W0*det*w[0][19];
    const double G565 = Jinv_00*Jinv_12*W0*det*w[0][23];
    const double G566 = Jinv_01*Jinv_10*W0*det*w[0][9];
    const double G567 = Jinv_01*Jinv_12*W0*det*w[0][11];
    const double G568 = Jinv_02*Jinv_10*W0*det*w[0][33];
    const double G569 = Jinv_02*Jinv_11*W0*det*w[0][31];
    const double G570 = Jinv_00*Jinv_21*W0*det*w[0][19];
    const double G571 = Jinv_00*Jinv_22*W0*det*w[0][23];
    const double G572 = Jinv_01*Jinv_20*W0*det*w[0][9];
    const double G573 = Jinv_01*Jinv_22*W0*det*w[0][11];
    const double G574 = Jinv_02*Jinv_20*W0*det*w[0][33];
    const double G575 = Jinv_02*Jinv_21*W0*det*w[0][31];
    const double G576 = Jinv_00*Jinv_21*W0*det*w[0][1];
    const double G577 = Jinv_00*Jinv_22*W0*det*w[0][5];
    const double G578 = Jinv_01*Jinv_22*W0*det*w[0][23];
    const double G579 = Jinv_02*Jinv_20*W0*det*w[0][27];
    const double G580 = Jinv_02*Jinv_21*W0*det*w[0][25];
    const double G581 = Jinv_10*Jinv_21*W0*det*w[0][22];
    const double G582 = Jinv_11*Jinv_20*W0*det*w[0][5];
    const double G583 = Jinv_11*Jinv_22*W0*det*w[0][29];
    const double G584 = Jinv_12*Jinv_20*W0*det*w[0][2];
    const double G585 = Jinv_12*Jinv_21*W0*det*w[0][20];
    const double G586 = Jinv_00*Jinv_11*W0*det*w[0][6];
    const double G587 = Jinv_00*Jinv_12*W0*det*w[0][30];
    const double G588 = Jinv_01*Jinv_12*W0*det*w[0][33];
    const double G589 = Jinv_02*Jinv_10*W0*det*w[0][22];
    const double G590 = Jinv_02*Jinv_11*W0*det*w[0][10];
    const double G591 = Jinv_00*Jinv_11*W0*det*w[0][5];
    const double G592 = Jinv_00*Jinv_12*W0*det*w[0][2];
    const double G593 = Jinv_01*Jinv_10*W0*det*w[0][22];
    const double G594 = Jinv_01*Jinv_12*W0*det*w[0][20];
    const double G595 = Jinv_02*Jinv_11*W0*det*w[0][29];
    const double G596 = Jinv_10*Jinv_10*W0*det*w[0][4];
    const double G597 = Jinv_10*Jinv_11*W0*det*w[0][22];
    const double G598 = Jinv_10*Jinv_11*W0*det*w[0][5];
    const double G599 = Jinv_10*Jinv_12*W0*det*w[0][2];
    const double G600 = Jinv_11*Jinv_11*W0*det*w[0][23];
    const double G601 = Jinv_11*Jinv_12*W0*det*w[0][20];
    const double G602 = Jinv_11*Jinv_12*W0*det*w[0][29];
    const double G603 = Jinv_12*Jinv_12*W0*det*w[0][26];
    const double G604 = Jinv_20*Jinv_20*W0*det*w[0][3];
    const double G605 = Jinv_20*Jinv_21*W0*det*w[0][1];
    const double G606 = Jinv_20*Jinv_22*W0*det*w[0][27];
    const double G607 = Jinv_20*Jinv_22*W0*det*w[0][5];
    const double G608 = Jinv_21*Jinv_21*W0*det*w[0][19];
    const double G609 = Jinv_21*Jinv_22*W0*det*w[0][23];
    const double G610 = Jinv_21*Jinv_22*W0*det*w[0][25];
    const double G611 = Jinv_22*Jinv_22*W0*det*w[0][29];
    const double G612 = Jinv_10*Jinv_10*W0*det*w[0][21];
    const double G613 = Jinv_10*Jinv_11*W0*det*w[0][19];
    const double G614 = Jinv_10*Jinv_11*W0*det*w[0][9];
    const double G615 = Jinv_10*Jinv_12*W0*det*w[0][23];
    const double G616 = Jinv_10*Jinv_12*W0*det*w[0][33];
    const double G617 = Jinv_11*Jinv_11*W0*det*w[0][7];
    const double G618 = Jinv_11*Jinv_12*W0*det*w[0][11];
    const double G619 = Jinv_11*Jinv_12*W0*det*w[0][31];
    const double G620 = Jinv_12*Jinv_12*W0*det*w[0][35];
    
    // Loop quadrature points (tensor/monomial terms (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80))
    // Number of operations to compute element tensor for following IP loop = 1620
    // Only 1 integration point, omitting IP loop.
    
    // Number of operations to compute declarations = 648
    const double Gip0 = G0 + G1 + G2 + G3 + G4 + G5 + G6 + G7 + G8;
    const double Gip1 = G503 + G596 + G597 + G598 + G599 + G600 + G601 + G602 + G603;
    const double Gip2 = G18 + G19 + G20 + G21 + G22 + G23 + G24 + G25 + G26;
    const double Gip3 = G27 + G28 + G29 + G30 + G31 + G32 + G33 + G34 + G35;
    const double Gip4 = G45 + G46 + G47 + G48 + G49 + G50 + G51 + G52 + G53;
    const double Gip5 = G228 + G232 + G236 + G367 + G419 + G420 + G421 + G422 + G423;
    const double Gip6 = G250 + G326 + G327 + G328 + G329 + G330 + G331 + G332 + G333;
    const double Gip7 = G72 + G73 + G74 + G75 + G76 + G77 + G78 + G79 + G80;
    const double Gip8 = G209 + G213 + G216 + G222 + G223 + G224 + G225 + G226 + G227;
    const double Gip9 = G90 + G91 + G92 + G93 + G94 + G95 + G96 + G97 + G98;
    const double Gip10 = G45 + G49 + G53 + G542 + G543 + G544 + G545 + G546 + G547;
    const double Gip11 = G36 + G37 + G38 + G39 + G40 + G41 + G42 + G43 + G44;
    const double Gip12 = G117 + G118 + G119 + G120 + G121 + G32 + G81 + G85 + G89;
    const double Gip13 = G334 + G335 + G336 + G337 + G338 + G339 + G340 + G341 + G69;
    const double Gip14 = G131 + G132 + G133 + G134 + G135 + G136 + G137 + G138 + G139;
    const double Gip15 = G13 + G149 + G150 + G151 + G152 + G153 + G154 + G17 + G9;
    const double Gip16 = G155 + G156 + G157 + G158 + G159 + G160 + G161 + G162 + G163;
    const double Gip17 = G164 + G165 + G166 + G167 + G168 + G169 + G170 + G171 + G172;
    const double Gip18 = G612 + G613 + G614 + G615 + G616 + G617 + G618 + G619 + G620;
    const double Gip19 = G518 + G519 + G520 + G521 + G522 + G523 + G524 + G525 + G526;
    const double Gip20 = G182 + G183 + G184 + G185 + G186 + G187 + G188 + G189 + G190;
    const double Gip21 = G191 + G192 + G193 + G194 + G195 + G196 + G197 + G198 + G199;
    const double Gip22 = G200 + G201 + G202 + G203 + G204 + G205 + G206 + G207 + G208;
    const double Gip23 = G136 + G209 + G210 + G211 + G212 + G213 + G214 + G215 + G216;
    const double Gip24 = G131 + G135 + G139 + G217 + G218 + G219 + G220 + G221 + G222;
    const double Gip25 = G228 + G229 + G230 + G231 + G232 + G233 + G234 + G235 + G236;
    const double Gip26 = G237 + G238 + G239 + G240 + G241 + G242 + G243 + G244 + G245;
    const double Gip27 = G246 + G247 + G248 + G249 + G250 + G251 + G252 + G253 + G254;
    const double Gip28 = G255 + G256 + G257 + G258 + G259 + G260 + G90 + G94 + G98;
    const double Gip29 = G319 + G356 + G357 + G358 + G359 + G360 + G361 + G362 + G363;
    const double Gip30 = G269 + G27 + G270 + G271 + G272 + G273 + G31 + G35 + G88;
    const double Gip31 = G274 + G275 + G276 + G277 + G278 + G279 + G280 + G281 + G282;
    const double Gip32 = G238 + G283 + G284 + G285 + G286 + G287 + G288 + G289 + G290;
    const double Gip33 = G140 + G141 + G142 + G143 + G144 + G145 + G146 + G147 + G148;
    const double Gip34 = G103 + G107 + G297 + G298 + G299 + G300 + G301 + G302 + G99;
    const double Gip35 = G356 + G359 + G363 + G467 + G591 + G592 + G593 + G594 + G595;
    const double Gip36 = G274 + G278 + G282 + G311 + G312 + G313 + G314 + G315 + G316;
    const double Gip37 = G317 + G318 + G319 + G320 + G321 + G322 + G323 + G324 + G325;
    const double Gip38 = G63 + G64 + G65 + G66 + G67 + G68 + G69 + G70 + G71;
    const double Gip39 = G122 + G123 + G124 + G125 + G126 + G127 + G128 + G129 + G130;
    const double Gip40 = G0 + G342 + G343 + G344 + G345 + G346 + G347 + G4 + G8;
    const double Gip41 = G348 + G349 + G350 + G351 + G352 + G353 + G354 + G355 + G55;
    const double Gip42 = G110 + G261 + G262 + G263 + G264 + G265 + G266 + G267 + G268;
    const double Gip43 = G364 + G365 + G366 + G367 + G368 + G369 + G370 + G371 + G372;
    const double Gip44 = G373 + G374 + G375 + G376 + G377 + G378 + G379 + G380 + G381;
    const double Gip45 = G382 + G383 + G384 + G385 + G386 + G387 + G388 + G389 + G390;
    const double Gip46 = G229 + G364 + G368 + G372 + G391 + G392 + G393 + G394 + G395;
    const double Gip47 = G396 + G397 + G398 + G399 + G400 + G401 + G402 + G403 + G404;
    const double Gip48 = G405 + G406 + G407 + G408 + G409 + G410 + G411 + G412 + G413;
    const double Gip49 = G129 + G405 + G409 + G413 + G414 + G415 + G416 + G417 + G418;
    const double Gip50 = G54 + G55 + G56 + G57 + G58 + G59 + G60 + G61 + G62;
    const double Gip51 = G424 + G425 + G426 + G427 + G428 + G429 + G430 + G431 + G432;
    const double Gip52 = G433 + G434 + G435 + G436 + G437 + G438 + G439 + G440 + G441;
    const double Gip53 = G396 + G400 + G404 + G442 + G443 + G444 + G445 + G446 + G447;
    const double Gip54 = G424 + G428 + G432 + G448 + G449 + G450 + G451 + G452 + G453;
    const double Gip55 = G454 + G455 + G456 + G457 + G458 + G459 + G460 + G461 + G462;
    const double Gip56 = G317 + G321 + G325 + G463 + G464 + G465 + G466 + G467 + G468;
    const double Gip57 = G348 + G351 + G355 + G529 + G576 + G577 + G578 + G579 + G580;
    const double Gip58 = G469 + G470 + G471 + G472 + G473 + G474 + G63 + G67 + G71;
    const double Gip59 = G188 + G475 + G476 + G477 + G478 + G479 + G480 + G481 + G482;
    const double Gip60 = G100 + G101 + G102 + G103 + G104 + G105 + G106 + G107 + G99;
    const double Gip61 = G291 + G292 + G293 + G294 + G295 + G296 + G72 + G76 + G80;
    const double Gip62 = G499 + G500 + G501 + G502 + G503 + G504 + G505 + G506 + G507;
    const double Gip63 = G122 + G126 + G130 + G410 + G508 + G509 + G510 + G511 + G512;
    const double Gip64 = G334 + G338 + G341 + G470 + G513 + G514 + G515 + G516 + G517;
    const double Gip65 = G81 + G82 + G83 + G84 + G85 + G86 + G87 + G88 + G89;
    const double Gip66 = G527 + G528 + G529 + G530 + G531 + G532 + G54 + G58 + G62;
    const double Gip67 = G15 + G556 + G557 + G558 + G559 + G560 + G561 + G562 + G563;
    const double Gip68 = G533 + G534 + G535 + G536 + G537 + G538 + G539 + G540 + G541;
    const double Gip69 = G388 + G483 + G484 + G485 + G486 + G487 + G488 + G489 + G490;
    const double Gip70 = G460 + G548 + G549 + G550 + G551 + G552 + G553 + G554 + G555;
    const double Gip71 = G173 + G174 + G175 + G176 + G177 + G178 + G179 + G180 + G181;
    const double Gip72 = G140 + G144 + G148 + G564 + G565 + G566 + G567 + G568 + G569;
    const double Gip73 = G533 + G537 + G541 + G570 + G571 + G572 + G573 + G574 + G575;
    const double Gip74 = G194 + G303 + G304 + G305 + G306 + G307 + G308 + G309 + G310;
    const double Gip75 = G150 + G556 + G560 + G563 + G581 + G582 + G583 + G584 + G585;
    const double Gip76 = G293 + G491 + G494 + G498 + G586 + G587 + G588 + G589 + G590;
    const double Gip77 = G10 + G11 + G12 + G13 + G14 + G15 + G16 + G17 + G9;
    const double Gip78 = G491 + G492 + G493 + G494 + G495 + G496 + G497 + G498 + G73;
    const double Gip79 = G434 + G604 + G605 + G606 + G607 + G608 + G609 + G610 + G611;
    const double Gip80 = G108 + G109 + G110 + G111 + G112 + G113 + G114 + G115 + G116;
    
    // Loop primary indices.
    // Number of operations for primary indices = 972
    for (unsigned int j = 0; j < 2; j++)
    {
      for (unsigned int k = 0; k < 2; k++)
      {
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip0;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip1;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip2;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip3;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip4;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip5;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip6;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip7;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip8;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip9;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip10;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip11;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip12;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip13;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip14;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip15;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip16;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip17;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip18;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip19;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip20;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip21;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip22;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip23;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip24;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip25;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip26;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip27;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip28;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip29;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip30;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip31;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip32;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip33;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip34;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip35;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip36;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip37;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip38;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip39;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip40;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip41;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip42;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip43;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip44;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip45;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip46;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip47;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip48;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip49;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip50;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip51;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip52;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip53;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip54;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip55;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip56;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip57;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip58;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip59;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip60;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip61;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip62;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip63;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc2[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip64;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip65;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip66;
        // Number of operations to compute entry = 3
        A[nzc6[j]*12 + nzc0[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip67;
        // Number of operations to compute entry = 3
        A[nzc4[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip68;
        // Number of operations to compute entry = 3
        A[nzc1[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip69;
        // Number of operations to compute entry = 3
        A[nzc2[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip70;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip71;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc8[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip72;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip73;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc3[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip74;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc1[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip75;
        // Number of operations to compute entry = 3
        A[nzc8[j]*12 + nzc5[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip76;
        // Number of operations to compute entry = 3
        A[nzc0[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip77;
        // Number of operations to compute entry = 3
        A[nzc7[j]*12 + nzc6[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip78;
        // Number of operations to compute entry = 3
        A[nzc3[j]*12 + nzc4[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip79;
        // Number of operations to compute entry = 3
        A[nzc5[j]*12 + nzc7[k]] += P_t2_p1_s2[0][j]*P_t2_p1_s2[0][k]*Gip80;
      }// end loop over 'k'
    }// end loop over 'j'
    
}

/// Constructor
UFC_Plas3DBilinearForm::UFC_Plas3DBilinearForm() : ufc::form()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DBilinearForm::~UFC_Plas3DBilinearForm()
{
    // Do nothing
}

/// Return a string identifying the form
const char* UFC_Plas3DBilinearForm::signature() const
{
    return "w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][0]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][1]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][2]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][3]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][3]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][4]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][4]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][5]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][5]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][6]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][7]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][8]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][9]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][9]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][10]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][10]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][11]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][11]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][12]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][13]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][14]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][15]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][15]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][16]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][16]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][17]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][17]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][18]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][19]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][20]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][21]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][21]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][22]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][22]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][23]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][23]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][18]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][19]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][20]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][21]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][21]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][22]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][22]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][23]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][23]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][24]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][25]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][26]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][27]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][27]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][28]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][28]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][29]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][29]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][24]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][25]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][26]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][27]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][27]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][28]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][28]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][29]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx0)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][29]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][30]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][31]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][32]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][33]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][33]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][34]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][34]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][35]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx2)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][35]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][30]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][31]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][32]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][33]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][33]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][34]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][34]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][35]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*dX(0) + w0_a1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35](dXa0[0, 1, 2]/dx1)(dXa2[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35][35]*((d/dXa2[0, 1, 2])vi1[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*dX(0)";
}

/// Return the rank of the global tensor (r)
unsigned int UFC_Plas3DBilinearForm::rank() const
{
    return 2;
}

/// Return the number of coefficients (n)
unsigned int UFC_Plas3DBilinearForm::num_coefficients() const
{
    return 1;
}

/// Return the number of cell integrals
unsigned int UFC_Plas3DBilinearForm::num_cell_integrals() const
{
    return 1;
}
  
/// Return the number of exterior facet integrals
unsigned int UFC_Plas3DBilinearForm::num_exterior_facet_integrals() const
{
    return 0;
}
  
/// Return the number of interior facet integrals
unsigned int UFC_Plas3DBilinearForm::num_interior_facet_integrals() const
{
    return 0;
}

/// Create a new finite element for argument function i
ufc::finite_element* UFC_Plas3DBilinearForm::create_finite_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_finite_element_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_finite_element_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_finite_element_2();
      break;
    }
    return 0;
}

/// Create a new dof map for argument function i
ufc::dof_map* UFC_Plas3DBilinearForm::create_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DBilinearForm_dof_map_0();
      break;
    case 1:
      return new UFC_Plas3DBilinearForm_dof_map_1();
      break;
    case 2:
      return new UFC_Plas3DBilinearForm_dof_map_2();
      break;
    }
    return 0;
}

/// Create a new cell integral on sub domain i
ufc::cell_integral* UFC_Plas3DBilinearForm::create_cell_integral(unsigned int i) const
{
    return new UFC_Plas3DBilinearForm_cell_integral_0();
}

/// Create a new exterior facet integral on sub domain i
ufc::exterior_facet_integral* UFC_Plas3DBilinearForm::create_exterior_facet_integral(unsigned int i) const
{
    return 0;
}

/// Create a new interior facet integral on sub domain i
ufc::interior_facet_integral* UFC_Plas3DBilinearForm::create_interior_facet_integral(unsigned int i) const
{
    return 0;
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_0_0::UFC_Plas3DLinearForm_finite_element_0_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_0_0::~UFC_Plas3DLinearForm_finite_element_0_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_0_0::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_0_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_0_0::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_0_0::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_0_0::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_0_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_0_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_0_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_0_0::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_0_0::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_0_0();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_0_1::UFC_Plas3DLinearForm_finite_element_0_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_0_1::~UFC_Plas3DLinearForm_finite_element_0_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_0_1::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_0_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_0_1::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_0_1::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_0_1::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_0_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_0_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_0_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_0_1::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_0_1::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_0_1();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_0_2::UFC_Plas3DLinearForm_finite_element_0_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_0_2::~UFC_Plas3DLinearForm_finite_element_0_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_0_2::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_0_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_0_2::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_0_2::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_0_2::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_0_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_0_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_0_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_0_2::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_0_2::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_0_2();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_0::UFC_Plas3DLinearForm_finite_element_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_0::~UFC_Plas3DLinearForm_finite_element_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_0::signature() const
{
    return "Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_0::space_dimension() const
{
    return 12;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_0::value_rank() const
{
    return 1;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_0::value_dimension(unsigned int i) const
{
    return 3;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[12][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[12][1] = {{1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}};
    const static double D[12][1][3] = {{{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[3];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 3; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[3] = dof_values[1];
    vertex_values[6] = dof_values[2];
    vertex_values[9] = dof_values[3];
    // Evaluate at vertices and use affine mapping
    vertex_values[1] = dof_values[4];
    vertex_values[4] = dof_values[5];
    vertex_values[7] = dof_values[6];
    vertex_values[10] = dof_values[7];
    // Evaluate at vertices and use affine mapping
    vertex_values[2] = dof_values[8];
    vertex_values[5] = dof_values[9];
    vertex_values[8] = dof_values[10];
    vertex_values[11] = dof_values[11];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_0::num_sub_elements() const
{
    return 3;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_0::create_sub_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_finite_element_0_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_finite_element_0_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_finite_element_0_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_1_0::UFC_Plas3DLinearForm_finite_element_1_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_1_0::~UFC_Plas3DLinearForm_finite_element_1_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_1_0::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_1_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_1_0::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_1_0::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_1_0::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_1_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_1_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_1_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_1_0::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_1_0::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_1_0();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_1_1::UFC_Plas3DLinearForm_finite_element_1_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_1_1::~UFC_Plas3DLinearForm_finite_element_1_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_1_1::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_1_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_1_1::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_1_1::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_1_1::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_1_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_1_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_1_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_1_1::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_1_1::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_1_1();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_1_2::UFC_Plas3DLinearForm_finite_element_1_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_1_2::~UFC_Plas3DLinearForm_finite_element_1_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_1_2::signature() const
{
    return "Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_1_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_1_2::space_dimension() const
{
    return 4;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_1_2::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_1_2::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_1_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[4][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[4][1] = {{1}, {1}, {1}, {1}};
    const static double D[4][1][1] = {{{1}}, {{1}}, {{1}}, {{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_1_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_1_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[1] = dof_values[1];
    vertex_values[2] = dof_values[2];
    vertex_values[3] = dof_values[3];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_1_2::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_1_2::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_1_2();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_1::UFC_Plas3DLinearForm_finite_element_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_1::~UFC_Plas3DLinearForm_finite_element_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_1::signature() const
{
    return "Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_1::space_dimension() const
{
    return 12;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_1::value_rank() const
{
    return 1;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_1::value_dimension(unsigned int i) const
{
    return 3;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis() is not yet implemented.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("The vectorised version of evaluate_basis_derivatives() is not yet implemented.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[12][1][3] = {{{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 0, 1}}};
    const static double W[12][1] = {{1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}, {1}};
    const static double D[12][1][3] = {{{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{1, 0, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 1, 0}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}, {{0, 0, 1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[3];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 3; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    // Evaluate at vertices and use affine mapping
    vertex_values[0] = dof_values[0];
    vertex_values[3] = dof_values[1];
    vertex_values[6] = dof_values[2];
    vertex_values[9] = dof_values[3];
    // Evaluate at vertices and use affine mapping
    vertex_values[1] = dof_values[4];
    vertex_values[4] = dof_values[5];
    vertex_values[7] = dof_values[6];
    vertex_values[10] = dof_values[7];
    // Evaluate at vertices and use affine mapping
    vertex_values[2] = dof_values[8];
    vertex_values[5] = dof_values[9];
    vertex_values[8] = dof_values[10];
    vertex_values[11] = dof_values[11];
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_1::num_sub_elements() const
{
    return 3;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_1::create_sub_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_finite_element_1_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_finite_element_1_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_finite_element_1_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2_0::UFC_Plas3DLinearForm_finite_element_2_0() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2_0::~UFC_Plas3DLinearForm_finite_element_2_0()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2_0::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2_0::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2_0::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2_0::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2_0::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_0::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_0::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_0::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_0::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2_0::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2_0::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2_0::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2_0::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2_0::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_2_0();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2_1::UFC_Plas3DLinearForm_finite_element_2_1() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2_1::~UFC_Plas3DLinearForm_finite_element_2_1()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2_1::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2_1::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2_1::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2_1::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2_1::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_1::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_1::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_1::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_1::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2_1::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2_1::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2_1::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2_1::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2_1::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_2_1();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2_2::UFC_Plas3DLinearForm_finite_element_2_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2_2::~UFC_Plas3DLinearForm_finite_element_2_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2_2::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2_2::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2_2::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2_2::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2_2::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2_2::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_2_2();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2_3::UFC_Plas3DLinearForm_finite_element_2_3() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2_3::~UFC_Plas3DLinearForm_finite_element_2_3()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2_3::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2_3::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2_3::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2_3::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2_3::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_3::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_3::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_3::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_3::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2_3::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2_3::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2_3::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2_3::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2_3::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_2_3();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2_4::UFC_Plas3DLinearForm_finite_element_2_4() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2_4::~UFC_Plas3DLinearForm_finite_element_2_4()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2_4::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2_4::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2_4::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2_4::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2_4::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_4::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_4::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_4::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_4::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2_4::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2_4::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2_4::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2_4::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2_4::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_2_4();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2_5::UFC_Plas3DLinearForm_finite_element_2_5() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2_5::~UFC_Plas3DLinearForm_finite_element_2_5()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2_5::signature() const
{
    return "Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2_5::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2_5::space_dimension() const
{
    return 1;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2_5::value_rank() const
{
    return 0;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2_5::value_dimension(unsigned int i) const
{
    return 1;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_5::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_5::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_5::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2_5::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2_5::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[1][1][3] = {{{0.25, 0.25, 0.25}}};
    const static double W[1][1] = {{1}};
    const static double D[1][1][1] = {{{1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[1];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 1; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2_5::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2_5::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2_5::num_sub_elements() const
{
    return 1;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2_5::create_sub_element(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_finite_element_2_5();
}


/// Constructor
UFC_Plas3DLinearForm_finite_element_2::UFC_Plas3DLinearForm_finite_element_2() : ufc::finite_element()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_finite_element_2::~UFC_Plas3DLinearForm_finite_element_2()
{
    // Do nothing
}

/// Return a string identifying the finite element
const char* UFC_Plas3DLinearForm_finite_element_2::signature() const
{
    return "Mixed finite element: [Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron]";
}

/// Return the cell shape
ufc::shape UFC_Plas3DLinearForm_finite_element_2::cell_shape() const
{
    return ufc::tetrahedron;
}

/// Return the dimension of the finite element function space
unsigned int UFC_Plas3DLinearForm_finite_element_2::space_dimension() const
{
    return 6;
}

/// Return the rank of the value space
unsigned int UFC_Plas3DLinearForm_finite_element_2::value_rank() const
{
    return 1;
}

/// Return the dimension of the value space for axis i
unsigned int UFC_Plas3DLinearForm_finite_element_2::value_dimension(unsigned int i) const
{
    return 6;
}

/// Evaluate basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2::evaluate_basis(unsigned int i,
                                   double* values,
                                   const double* coordinates,
                                   const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis not generated (compiled with -fno-evaluate_basis)");
}

/// Evaluate all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2::evaluate_basis_all(double* values,
                                       const double* coordinates,
                                       const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_all() is not supported for QuadratureElement.");
}

/// Evaluate order n derivatives of basis function i at given point in cell
void UFC_Plas3DLinearForm_finite_element_2::evaluate_basis_derivatives(unsigned int i,
                                               unsigned int n,
                                               double* values,
                                               const double* coordinates,
                                               const ufc::cell& c) const
{
    throw std::runtime_error("// Function evaluate_basis_derivatives not generated (compiled with -fno-evaluate_basis_derivatives)");
}

/// Evaluate order n derivatives of all basis functions at given point in cell
void UFC_Plas3DLinearForm_finite_element_2::evaluate_basis_derivatives_all(unsigned int n,
                                                   double* values,
                                                   const double* coordinates,
                                                   const ufc::cell& c) const
{
    throw std::runtime_error("evaluate_basis_derivatives_all() is not supported for QuadratureElement.");
}

/// Evaluate linear functional for dof i on the function f
double UFC_Plas3DLinearForm_finite_element_2::evaluate_dof(unsigned int i,
                                   const ufc::function& f,
                                   const ufc::cell& c) const
{
    // The reference points, direction and weights:
    const static double X[6][1][3] = {{{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}, {{0.25, 0.25, 0.25}}};
    const static double W[6][1] = {{1}, {1}, {1}, {1}, {1}, {1}};
    const static double D[6][1][6] = {{{1, 0, 0, 0, 0, 0}}, {{0, 1, 0, 0, 0, 0}}, {{0, 0, 1, 0, 0, 0}}, {{0, 0, 0, 1, 0, 0}}, {{0, 0, 0, 0, 1, 0}}, {{0, 0, 0, 0, 0, 1}}};
    
    const double * const * x = c.coordinates;
    double result = 0.0;
    // Iterate over the points:
    // Evaluate basis functions for affine mapping
    const double w0 = 1.0 - X[i][0][0] - X[i][0][1] - X[i][0][2];
    const double w1 = X[i][0][0];
    const double w2 = X[i][0][1];
    const double w3 = X[i][0][2];
    
    // Compute affine mapping y = F(X)
    double y[3];
    y[0] = w0*x[0][0] + w1*x[1][0] + w2*x[2][0] + w3*x[3][0];
    y[1] = w0*x[0][1] + w1*x[1][1] + w2*x[2][1] + w3*x[3][1];
    y[2] = w0*x[0][2] + w1*x[1][2] + w2*x[2][2] + w3*x[3][2];
    
    // Evaluate function at physical points
    double values[6];
    f.evaluate(values, y, c);
    
    // Map function values using appropriate mapping
    // Affine map: Do nothing
    
    // Note that we do not map the weights (yet).
    
    // Take directional components
    for(int k = 0; k < 6; k++)
      result += values[k]*D[i][0][k];
    // Multiply by weights 
    result *= W[i][0];
    
    return result;
}

/// Evaluate linear functionals for all dofs on the function f
void UFC_Plas3DLinearForm_finite_element_2::evaluate_dofs(double* values,
                                  const ufc::function& f,
                                  const ufc::cell& c) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Interpolate vertex values from dof values
void UFC_Plas3DLinearForm_finite_element_2::interpolate_vertex_values(double* vertex_values,
                                              const double* dof_values,
                                              const ufc::cell& c) const
{
    throw std::runtime_error("interpolate_vertex_values() is not supported for QuadratureElement");
}

/// Return the number of sub elements (for a mixed element)
unsigned int UFC_Plas3DLinearForm_finite_element_2::num_sub_elements() const
{
    return 6;
}

/// Create a new finite element for sub element i (for a mixed element)
ufc::finite_element* UFC_Plas3DLinearForm_finite_element_2::create_sub_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_finite_element_2_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_finite_element_2_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_finite_element_2_2();
      break;
    case 3:
      return new UFC_Plas3DLinearForm_finite_element_2_3();
      break;
    case 4:
      return new UFC_Plas3DLinearForm_finite_element_2_4();
      break;
    case 5:
      return new UFC_Plas3DLinearForm_finite_element_2_5();
      break;
    }
    return 0;
}

/// Constructor
UFC_Plas3DLinearForm_dof_map_0_0::UFC_Plas3DLinearForm_dof_map_0_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_0_0::~UFC_Plas3DLinearForm_dof_map_0_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_0_0::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_0_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_0_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_0_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_0_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0_0::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_0_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_0_0::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_0_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_0_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_0_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_0_0::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_0_0::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_0_0();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_0_1::UFC_Plas3DLinearForm_dof_map_0_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_0_1::~UFC_Plas3DLinearForm_dof_map_0_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_0_1::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_0_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_0_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_0_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_0_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0_1::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_0_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_0_1::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_0_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_0_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_0_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_0_1::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_0_1::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_0_1();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_0_2::UFC_Plas3DLinearForm_dof_map_0_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_0_2::~UFC_Plas3DLinearForm_dof_map_0_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_0_2::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_0_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_0_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_0_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_0_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0_2::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_0_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_0_2::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_0_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_0_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_0_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_0_2::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_0_2::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_0_2();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_0::UFC_Plas3DLinearForm_dof_map_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_0::~UFC_Plas3DLinearForm_dof_map_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_0::signature() const
{
    return "FFC dof map for Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = 3*m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_0::local_dimension() const
{
    return 12;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_0::num_facet_dofs() const
{
    return 9;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
    unsigned int offset = m.num_entities[0];
    dofs[4] = offset + c.entity_indices[0][0];
    dofs[5] = offset + c.entity_indices[0][1];
    dofs[6] = offset + c.entity_indices[0][2];
    dofs[7] = offset + c.entity_indices[0][3];
    offset = offset + m.num_entities[0];
    dofs[8] = offset + c.entity_indices[0][0];
    dofs[9] = offset + c.entity_indices[0][1];
    dofs[10] = offset + c.entity_indices[0][2];
    dofs[11] = offset + c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 5;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 9;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 11;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 6;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 10;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
    coordinates[4][0] = x[0][0];
    coordinates[4][1] = x[0][1];
    coordinates[4][2] = x[0][2];
    coordinates[5][0] = x[1][0];
    coordinates[5][1] = x[1][1];
    coordinates[5][2] = x[1][2];
    coordinates[6][0] = x[2][0];
    coordinates[6][1] = x[2][1];
    coordinates[6][2] = x[2][2];
    coordinates[7][0] = x[3][0];
    coordinates[7][1] = x[3][1];
    coordinates[7][2] = x[3][2];
    coordinates[8][0] = x[0][0];
    coordinates[8][1] = x[0][1];
    coordinates[8][2] = x[0][2];
    coordinates[9][0] = x[1][0];
    coordinates[9][1] = x[1][1];
    coordinates[9][2] = x[1][2];
    coordinates[10][0] = x[2][0];
    coordinates[10][1] = x[2][1];
    coordinates[10][2] = x[2][2];
    coordinates[11][0] = x[3][0];
    coordinates[11][1] = x[3][1];
    coordinates[11][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_0::num_sub_dof_maps() const
{
    return 3;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_0::create_sub_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_dof_map_0_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_dof_map_0_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_dof_map_0_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_1_0::UFC_Plas3DLinearForm_dof_map_1_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_1_0::~UFC_Plas3DLinearForm_dof_map_1_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_1_0::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_1_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_1_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_1_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_1_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1_0::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_1_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_1_0::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_1_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_1_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_1_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_1_0::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_1_0::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_1_0();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_1_1::UFC_Plas3DLinearForm_dof_map_1_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_1_1::~UFC_Plas3DLinearForm_dof_map_1_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_1_1::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_1_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_1_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_1_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_1_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1_1::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_1_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_1_1::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_1_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_1_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_1_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_1_1::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_1_1::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_1_1();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_1_2::UFC_Plas3DLinearForm_dof_map_1_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_1_2::~UFC_Plas3DLinearForm_dof_map_1_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_1_2::signature() const
{
    return "FFC dof map for Lagrange finite element of degree 1 on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_1_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_1_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_1_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_1_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1_2::local_dimension() const
{
    return 4;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_1_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_1_2::num_facet_dofs() const
{
    return 3;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_1_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_1_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_1_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_1_2::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_1_2::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_1_2();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_1::UFC_Plas3DLinearForm_dof_map_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_1::~UFC_Plas3DLinearForm_dof_map_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_1::signature() const
{
    return "FFC dof map for Mixed finite element: [Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron, Lagrange finite element of degree 1 on a tetrahedron]";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return true;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return false;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = 3*m.num_entities[0];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_1::local_dimension() const
{
    return 12;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_1::num_facet_dofs() const
{
    return 9;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[0][0];
    dofs[1] = c.entity_indices[0][1];
    dofs[2] = c.entity_indices[0][2];
    dofs[3] = c.entity_indices[0][3];
    unsigned int offset = m.num_entities[0];
    dofs[4] = offset + c.entity_indices[0][0];
    dofs[5] = offset + c.entity_indices[0][1];
    dofs[6] = offset + c.entity_indices[0][2];
    dofs[7] = offset + c.entity_indices[0][3];
    offset = offset + m.num_entities[0];
    dofs[8] = offset + c.entity_indices[0][0];
    dofs[9] = offset + c.entity_indices[0][1];
    dofs[10] = offset + c.entity_indices[0][2];
    dofs[11] = offset + c.entity_indices[0][3];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      dofs[0] = 1;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 5;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 9;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 1:
      dofs[0] = 0;
      dofs[1] = 2;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 6;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 10;
      dofs[8] = 11;
      break;
    case 2:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 3;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 7;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 11;
      break;
    case 3:
      dofs[0] = 0;
      dofs[1] = 1;
      dofs[2] = 2;
      dofs[3] = 4;
      dofs[4] = 5;
      dofs[5] = 6;
      dofs[6] = 8;
      dofs[7] = 9;
      dofs[8] = 10;
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = x[0][0];
    coordinates[0][1] = x[0][1];
    coordinates[0][2] = x[0][2];
    coordinates[1][0] = x[1][0];
    coordinates[1][1] = x[1][1];
    coordinates[1][2] = x[1][2];
    coordinates[2][0] = x[2][0];
    coordinates[2][1] = x[2][1];
    coordinates[2][2] = x[2][2];
    coordinates[3][0] = x[3][0];
    coordinates[3][1] = x[3][1];
    coordinates[3][2] = x[3][2];
    coordinates[4][0] = x[0][0];
    coordinates[4][1] = x[0][1];
    coordinates[4][2] = x[0][2];
    coordinates[5][0] = x[1][0];
    coordinates[5][1] = x[1][1];
    coordinates[5][2] = x[1][2];
    coordinates[6][0] = x[2][0];
    coordinates[6][1] = x[2][1];
    coordinates[6][2] = x[2][2];
    coordinates[7][0] = x[3][0];
    coordinates[7][1] = x[3][1];
    coordinates[7][2] = x[3][2];
    coordinates[8][0] = x[0][0];
    coordinates[8][1] = x[0][1];
    coordinates[8][2] = x[0][2];
    coordinates[9][0] = x[1][0];
    coordinates[9][1] = x[1][1];
    coordinates[9][2] = x[1][2];
    coordinates[10][0] = x[2][0];
    coordinates[10][1] = x[2][1];
    coordinates[10][2] = x[2][2];
    coordinates[11][0] = x[3][0];
    coordinates[11][1] = x[3][1];
    coordinates[11][2] = x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_1::num_sub_dof_maps() const
{
    return 3;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_1::create_sub_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_dof_map_1_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_dof_map_1_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_dof_map_1_2();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2_0::UFC_Plas3DLinearForm_dof_map_2_0() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2_0::~UFC_Plas3DLinearForm_dof_map_2_0()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2_0::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2_0::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2_0::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2_0::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2_0::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_0::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_0::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2_0::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2_0::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2_0::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_0::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2_0::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2_0::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_0::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2_0::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2_0::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_2_0();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2_1::UFC_Plas3DLinearForm_dof_map_2_1() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2_1::~UFC_Plas3DLinearForm_dof_map_2_1()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2_1::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2_1::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2_1::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2_1::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2_1::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_1::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_1::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2_1::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2_1::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2_1::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_1::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2_1::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2_1::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_1::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2_1::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2_1::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_2_1();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2_2::UFC_Plas3DLinearForm_dof_map_2_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2_2::~UFC_Plas3DLinearForm_dof_map_2_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2_2::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_2::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2_2::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2_2::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2_2::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_2_2();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2_3::UFC_Plas3DLinearForm_dof_map_2_3() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2_3::~UFC_Plas3DLinearForm_dof_map_2_3()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2_3::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2_3::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2_3::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2_3::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2_3::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_3::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_3::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2_3::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2_3::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2_3::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_3::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2_3::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2_3::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_3::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2_3::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2_3::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_2_3();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2_4::UFC_Plas3DLinearForm_dof_map_2_4() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2_4::~UFC_Plas3DLinearForm_dof_map_2_4()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2_4::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2_4::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2_4::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2_4::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2_4::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_4::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_4::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2_4::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2_4::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2_4::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_4::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2_4::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2_4::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_4::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2_4::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2_4::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_2_4();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2_5::UFC_Plas3DLinearForm_dof_map_2_5() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2_5::~UFC_Plas3DLinearForm_dof_map_2_5()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2_5::signature() const
{
    return "FFC dof map for Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2_5::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2_5::init_mesh(const ufc::mesh& m)
{
    __global_dimension = m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2_5::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2_5::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_5::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2_5::local_dimension() const
{
    return 1;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2_5::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2_5::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2_5::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_5::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2_5::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2_5::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2_5::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2_5::num_sub_dof_maps() const
{
    return 1;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2_5::create_sub_dof_map(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_dof_map_2_5();
}


/// Constructor
UFC_Plas3DLinearForm_dof_map_2::UFC_Plas3DLinearForm_dof_map_2() : ufc::dof_map()
{
    __global_dimension = 0;
}

/// Destructor
UFC_Plas3DLinearForm_dof_map_2::~UFC_Plas3DLinearForm_dof_map_2()
{
    // Do nothing
}

/// Return a string identifying the dof map
const char* UFC_Plas3DLinearForm_dof_map_2::signature() const
{
    return "FFC dof map for Mixed finite element: [Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron, Quadrature element with 1 quadrature point(s) (1x1x1) on a tetrahedron]";
}

/// Return true iff mesh entities of topological dimension d are needed
bool UFC_Plas3DLinearForm_dof_map_2::needs_mesh_entities(unsigned int d) const
{
    switch ( d )
    {
    case 0:
      return false;
      break;
    case 1:
      return false;
      break;
    case 2:
      return false;
      break;
    case 3:
      return true;
      break;
    }
    return false;
}

/// Initialize dof map for mesh (return true iff init_cell() is needed)
bool UFC_Plas3DLinearForm_dof_map_2::init_mesh(const ufc::mesh& m)
{
    __global_dimension = 6*m.num_entities[3];
    return false;
}

/// Initialize dof map for given cell
void UFC_Plas3DLinearForm_dof_map_2::init_cell(const ufc::mesh& m,
                              const ufc::cell& c)
{
    // Do nothing
}

/// Finish initialization of dof map for cells
void UFC_Plas3DLinearForm_dof_map_2::init_cell_finalize()
{
    // Do nothing
}

/// Return the dimension of the global finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2::global_dimension() const
{
    return __global_dimension;
}

/// Return the dimension of the local finite element function space
unsigned int UFC_Plas3DLinearForm_dof_map_2::local_dimension() const
{
    return 6;
}

// Return the geometric dimension of the coordinates this dof map provides
unsigned int UFC_Plas3DLinearForm_dof_map_2::geometric_dimension() const
{
    return 3;
}

/// Return the number of dofs on each cell facet
unsigned int UFC_Plas3DLinearForm_dof_map_2::num_facet_dofs() const
{
    return 0;
}

/// Return the number of dofs associated with each cell entity of dimension d
unsigned int UFC_Plas3DLinearForm_dof_map_2::num_entity_dofs(unsigned int d) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the local-to-global mapping of dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2::tabulate_dofs(unsigned int* dofs,
                                  const ufc::mesh& m,
                                  const ufc::cell& c) const
{
    dofs[0] = c.entity_indices[3][0];
    unsigned int offset = m.num_entities[3];
    dofs[1] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[2] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[3] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[4] = offset + c.entity_indices[3][0];
    offset = offset + m.num_entities[3];
    dofs[5] = offset + c.entity_indices[3][0];
}

/// Tabulate the local-to-local mapping from facet dofs to cell dofs
void UFC_Plas3DLinearForm_dof_map_2::tabulate_facet_dofs(unsigned int* dofs,
                                        unsigned int facet) const
{
    switch ( facet )
    {
    case 0:
      
      break;
    case 1:
      
      break;
    case 2:
      
      break;
    case 3:
      
      break;
    }
}

/// Tabulate the local-to-local mapping of dofs on entity (d, i)
void UFC_Plas3DLinearForm_dof_map_2::tabulate_entity_dofs(unsigned int* dofs,
                                  unsigned int d, unsigned int i) const
{
    throw std::runtime_error("Not implemented (introduced in UFC v1.1).");
}

/// Tabulate the coordinates of all dofs on a cell
void UFC_Plas3DLinearForm_dof_map_2::tabulate_coordinates(double** coordinates,
                                         const ufc::cell& c) const
{
    const double * const * x = c.coordinates;
    coordinates[0][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[0][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[0][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[1][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[1][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[1][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[2][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[2][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[2][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[3][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[3][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[3][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[4][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[4][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[4][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
    coordinates[5][0] = 0.25*x[0][0] + 0.25*x[1][0] + 0.25*x[2][0] + 0.25*x[3][0];
    coordinates[5][1] = 0.25*x[0][1] + 0.25*x[1][1] + 0.25*x[2][1] + 0.25*x[3][1];
    coordinates[5][2] = 0.25*x[0][2] + 0.25*x[1][2] + 0.25*x[2][2] + 0.25*x[3][2];
}

/// Return the number of sub dof maps (for a mixed element)
unsigned int UFC_Plas3DLinearForm_dof_map_2::num_sub_dof_maps() const
{
    return 6;
}

/// Create a new dof_map for sub dof map i (for a mixed element)
ufc::dof_map* UFC_Plas3DLinearForm_dof_map_2::create_sub_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_dof_map_2_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_dof_map_2_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_dof_map_2_2();
      break;
    case 3:
      return new UFC_Plas3DLinearForm_dof_map_2_3();
      break;
    case 4:
      return new UFC_Plas3DLinearForm_dof_map_2_4();
      break;
    case 5:
      return new UFC_Plas3DLinearForm_dof_map_2_5();
      break;
    }
    return 0;
}


/// Constructor
UFC_Plas3DLinearForm_cell_integral_0::UFC_Plas3DLinearForm_cell_integral_0() : ufc::cell_integral()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm_cell_integral_0::~UFC_Plas3DLinearForm_cell_integral_0()
{
    // Do nothing
}

/// Tabulate the tensor for the contribution from a local cell
void UFC_Plas3DLinearForm_cell_integral_0::tabulate_tensor(double* A,
                                    const double * const * w,
                                    const ufc::cell& c) const
{
    // Extract vertex coordinates
    const double * const * x = c.coordinates;
    
    // Compute Jacobian of affine map from reference cell
    const double J_00 = x[1][0] - x[0][0];
    const double J_01 = x[2][0] - x[0][0];
    const double J_02 = x[3][0] - x[0][0];
    const double J_10 = x[1][1] - x[0][1];
    const double J_11 = x[2][1] - x[0][1];
    const double J_12 = x[3][1] - x[0][1];
    const double J_20 = x[1][2] - x[0][2];
    const double J_21 = x[2][2] - x[0][2];
    const double J_22 = x[3][2] - x[0][2];
      
    // Compute sub determinants
    const double d_00 = J_11*J_22 - J_12*J_21;
    const double d_01 = J_12*J_20 - J_10*J_22;
    const double d_02 = J_10*J_21 - J_11*J_20;
    
    const double d_10 = J_02*J_21 - J_01*J_22;
    const double d_11 = J_00*J_22 - J_02*J_20;
    const double d_12 = J_01*J_20 - J_00*J_21;
    
    const double d_20 = J_01*J_12 - J_02*J_11;
    const double d_21 = J_02*J_10 - J_00*J_12;
    const double d_22 = J_00*J_11 - J_01*J_10;
      
    // Compute determinant of Jacobian
    double detJ = J_00*d_00 + J_10*d_10 + J_20*d_20;
      
    // Compute inverse of Jacobian
    const double Jinv_00 = d_00 / detJ;
    const double Jinv_01 = d_10 / detJ;
    const double Jinv_02 = d_20 / detJ;
    const double Jinv_10 = d_01 / detJ;
    const double Jinv_11 = d_11 / detJ;
    const double Jinv_12 = d_21 / detJ;
    const double Jinv_20 = d_02 / detJ;
    const double Jinv_21 = d_12 / detJ;
    const double Jinv_22 = d_22 / detJ;
    
    // Set scale factor
    const double det = std::abs(detJ);
    
    
    // Compute element tensor (using quadrature representation, optimisation level 2)
    // Total number of operations to compute element tensor (from this point): 567
    
    // Reset values of the element tensor block
    for (unsigned int j = 0; j < 12; j++)
    {
      A[j] = 0;
    }// end loop over 'j'
    
    // Array of quadrature weights (tensor/monomial term 0)
    const static double W0[8] = {0.0369798563588528, 0.0160270405984766, 0.021157006454524, 0.00916942992147972, 0.0369798563588528, 0.0160270405984766, 0.021157006454524, 0.00916942992147972};
    // Array of quadrature weights (tensor/monomial terms (1, 2, 3, 4, 5, 6, 7, 8, 9))
    const static double W1 = -0.166666666666666;
    
    const static double P_t7_p0_s2[1][2] = \
    {{-1, 1}};
    // Array of non-zero columns
    static const unsigned int nzc2[2] = {4, 5};
    // Array of non-zero columns
    static const unsigned int nzc3[2] = {4, 6};
    // Array of non-zero columns
    static const unsigned int nzc4[2] = {4, 7};
    // Array of non-zero columns
    static const unsigned int nzc11[2] = {0, 3};
    // Array of non-zero columns
    static const unsigned int nzc0[2] = {8, 11};
    // Array of non-zero columns
    static const unsigned int nzc10[2] = {8, 10};
    // Array of non-zero columns
    static const unsigned int nzc6[2] = {8, 9};
    // Array of non-zero columns
    static const unsigned int nzc5[2] = {0, 1};
    // Array of non-zero columns
    static const unsigned int nzc1[2] = {0, 2};
    
    const static double P_t0_p0_a0[8][4] = \
    {{0.584747563204894, 0.156682637336818, 0.136054976802846, 0.122514822655441},
    {0.303772764814708, 0.0813956670146703, 0.0706797241593969, 0.544151844011225},
    {0.245713325211713, 0.0658386870600444, 0.565933165072801, 0.122514822655441},
    {0.127646562120385, 0.0342027932367665, 0.293998800631623, 0.544151844011225},
    {0.156682637336818, 0.584747563204894, 0.136054976802846, 0.122514822655441},
    {0.0813956670146704, 0.303772764814708, 0.0706797241593969, 0.544151844011225},
    {0.0658386870600444, 0.245713325211713, 0.565933165072801, 0.122514822655441},
    {0.0342027932367665, 0.127646562120385, 0.293998800631623, 0.544151844011225}};
    // Array of non-zero columns
    static const unsigned int nzc7[4] = {0, 1, 2, 3};
    // Array of non-zero columns
    static const unsigned int nzc8[4] = {4, 5, 6, 7};
    // Array of non-zero columns
    static const unsigned int nzc9[4] = {8, 9, 10, 11};
    
    // Number of operations to compute geometry constants = 81
    const double G0 = Jinv_20*W1*det*w[1][3];
    const double G1 = Jinv_21*W1*det*w[1][1];
    const double G2 = Jinv_22*W1*det*w[1][5];
    const double G3 = Jinv_00*W1*det*w[1][4];
    const double G4 = Jinv_01*W1*det*w[1][5];
    const double G5 = Jinv_02*W1*det*w[1][2];
    const double G6 = Jinv_20*W1*det*w[1][4];
    const double G7 = Jinv_21*W1*det*w[1][5];
    const double G8 = Jinv_22*W1*det*w[1][2];
    const double G9 = Jinv_20*W1*det*w[1][0];
    const double G10 = Jinv_21*W1*det*w[1][3];
    const double G11 = Jinv_22*W1*det*w[1][4];
    const double G12 = Jinv_10*W1*det*w[1][0];
    const double G13 = Jinv_11*W1*det*w[1][3];
    const double G14 = Jinv_12*W1*det*w[1][4];
    const double G15 = Jinv_10*W1*det*w[1][3];
    const double G16 = Jinv_11*W1*det*w[1][1];
    const double G17 = Jinv_12*W1*det*w[1][5];
    const double G18 = Jinv_00*W1*det*w[1][3];
    const double G19 = Jinv_01*W1*det*w[1][1];
    const double G20 = Jinv_02*W1*det*w[1][5];
    const double G21 = Jinv_10*W1*det*w[1][4];
    const double G22 = Jinv_11*W1*det*w[1][5];
    const double G23 = Jinv_12*W1*det*w[1][2];
    const double G24 = Jinv_00*W1*det*w[1][0];
    const double G25 = Jinv_01*W1*det*w[1][3];
    const double G26 = Jinv_02*W1*det*w[1][4];
    
    // Loop quadrature points (tensor/monomial terms (0,))
    // Number of operations to compute element tensor for following IP loop = 432
    for (unsigned int ip = 0; ip < 8; ip++)
    {
      
      // Declare function values.
      double F0 = 0;
      double F1 = 0;
      double F2 = 0;
      
      // Compute function values.
      // Number of operations to compute values = 24
      for (unsigned int r = 0; r < 4; r++)
      {
        F0 += P_t0_p0_a0[ip][r]*w[0][nzc7[r]];
        F1 += P_t0_p0_a0[ip][r]*w[0][nzc9[r]];
        F2 += P_t0_p0_a0[ip][r]*w[0][nzc8[r]];
      }// end loop over 'r'
      
      // Number of operations to compute declarations = 6
      const double Gip0 = F0*W0[ip]*det;
      const double Gip1 = F1*W0[ip]*det;
      const double Gip2 = F2*W0[ip]*det;
      
      // Loop primary indices.
      // Number of operations for primary indices = 24
      for (unsigned int j = 0; j < 4; j++)
      {
        // Number of operations to compute entry = 2
        A[nzc7[j]] += P_t0_p0_a0[ip][j]*Gip0;
        // Number of operations to compute entry = 2
        A[nzc9[j]] += P_t0_p0_a0[ip][j]*Gip1;
        // Number of operations to compute entry = 2
        A[nzc8[j]] += P_t0_p0_a0[ip][j]*Gip2;
      }// end loop over 'j'
      
    }// end loop over 'ip'
    // Loop quadrature points (tensor/monomial terms (1, 2, 3, 4, 5, 6, 7, 8, 9))
    // Number of operations to compute element tensor for following IP loop = 54
    // Only 1 integration point, omitting IP loop.
    
    // Number of operations to compute declarations = 18
    const double Gip0 = G0 + G1 + G2;
    const double Gip1 = G3 + G4 + G5;
    const double Gip2 = G6 + G7 + G8;
    const double Gip3 = G10 + G11 + G9;
    const double Gip4 = G12 + G13 + G14;
    const double Gip5 = G15 + G16 + G17;
    const double Gip6 = G18 + G19 + G20;
    const double Gip7 = G21 + G22 + G23;
    const double Gip8 = G24 + G25 + G26;
    
    // Loop primary indices.
    // Number of operations for primary indices = 36
    for (unsigned int j = 0; j < 2; j++)
    {
      // Number of operations to compute entry = 2
      A[nzc4[j]] += P_t7_p0_s2[0][j]*Gip0;
      // Number of operations to compute entry = 2
      A[nzc6[j]] += P_t7_p0_s2[0][j]*Gip1;
      // Number of operations to compute entry = 2
      A[nzc0[j]] += P_t7_p0_s2[0][j]*Gip2;
      // Number of operations to compute entry = 2
      A[nzc11[j]] += P_t7_p0_s2[0][j]*Gip3;
      // Number of operations to compute entry = 2
      A[nzc1[j]] += P_t7_p0_s2[0][j]*Gip4;
      // Number of operations to compute entry = 2
      A[nzc3[j]] += P_t7_p0_s2[0][j]*Gip5;
      // Number of operations to compute entry = 2
      A[nzc2[j]] += P_t7_p0_s2[0][j]*Gip6;
      // Number of operations to compute entry = 2
      A[nzc10[j]] += P_t7_p0_s2[0][j]*Gip7;
      // Number of operations to compute entry = 2
      A[nzc5[j]] += P_t7_p0_s2[0][j]*Gip8;
    }// end loop over 'j'
    
}

/// Constructor
UFC_Plas3DLinearForm::UFC_Plas3DLinearForm() : ufc::form()
{
    // Do nothing
}

/// Destructor
UFC_Plas3DLinearForm::~UFC_Plas3DLinearForm()
{
    // Do nothing
}

/// Return a string identifying the form
const char* UFC_Plas3DLinearForm::signature() const
{
    return "w0_a0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11] | vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][b0[0, 1, 2]]*va0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][b0[0, 1, 2]]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5][0]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5][3]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][0])*va1[0, 1, 2, 3, 4, 5][4]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5][3]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5][1]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][1])*va1[0, 1, 2, 3, 4, 5][5]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx0) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5][4]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx1) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5][5]*dX(0) + -w1_a1[0, 1, 2, 3, 4, 5](dXa0[0, 1, 2]/dx2) | ((d/dXa0[0, 1, 2])vi0[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11][2])*va1[0, 1, 2, 3, 4, 5][2]*dX(0)";
}

/// Return the rank of the global tensor (r)
unsigned int UFC_Plas3DLinearForm::rank() const
{
    return 1;
}

/// Return the number of coefficients (n)
unsigned int UFC_Plas3DLinearForm::num_coefficients() const
{
    return 2;
}

/// Return the number of cell integrals
unsigned int UFC_Plas3DLinearForm::num_cell_integrals() const
{
    return 1;
}
  
/// Return the number of exterior facet integrals
unsigned int UFC_Plas3DLinearForm::num_exterior_facet_integrals() const
{
    return 0;
}
  
/// Return the number of interior facet integrals
unsigned int UFC_Plas3DLinearForm::num_interior_facet_integrals() const
{
    return 0;
}

/// Create a new finite element for argument function i
ufc::finite_element* UFC_Plas3DLinearForm::create_finite_element(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_finite_element_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_finite_element_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_finite_element_2();
      break;
    }
    return 0;
}

/// Create a new dof map for argument function i
ufc::dof_map* UFC_Plas3DLinearForm::create_dof_map(unsigned int i) const
{
    switch ( i )
    {
    case 0:
      return new UFC_Plas3DLinearForm_dof_map_0();
      break;
    case 1:
      return new UFC_Plas3DLinearForm_dof_map_1();
      break;
    case 2:
      return new UFC_Plas3DLinearForm_dof_map_2();
      break;
    }
    return 0;
}

/// Create a new cell integral on sub domain i
ufc::cell_integral* UFC_Plas3DLinearForm::create_cell_integral(unsigned int i) const
{
    return new UFC_Plas3DLinearForm_cell_integral_0();
}

/// Create a new exterior facet integral on sub domain i
ufc::exterior_facet_integral* UFC_Plas3DLinearForm::create_exterior_facet_integral(unsigned int i) const
{
    return 0;
}

/// Create a new interior facet integral on sub domain i
ufc::interior_facet_integral* UFC_Plas3DLinearForm::create_interior_facet_integral(unsigned int i) const
{
    return 0;
}

